# Copyright (c) 2014 IBM Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for the IBM GPFS driver module."""

import re
import socket
from unittest import mock

import ddt
from oslo_config import cfg

from manila import context
from manila import exception
import manila.share.configuration as config
import manila.share.drivers.ibm.gpfs as gpfs
from manila.share import share_types
from manila import test
from manila.tests import fake_share
from manila import utils


CONF = cfg.CONF


@ddt.ddt
class GPFSShareDriverTestCase(test.TestCase):
    """Tests GPFSShareDriver."""

    def setUp(self):
        super(GPFSShareDriverTestCase, self).setUp()
        self._context = context.get_admin_context()
        self._gpfs_execute = mock.Mock(return_value=('', ''))
        self.GPFS_PATH = '/usr/lpp/mmfs/bin/'

        self._helper_fake = mock.Mock()
        CONF.set_default('driver_handles_share_servers', False)
        CONF.set_default('share_backend_name', 'GPFS')
        self.fake_conf = config.Configuration(None)
        self._driver = gpfs.GPFSShareDriver(execute=self._gpfs_execute,
                                            configuration=self.fake_conf)
        self._knfs_helper = gpfs.KNFSHelper(self._gpfs_execute,
                                            self.fake_conf)
        self._ces_helper = gpfs.CESHelper(self._gpfs_execute,
                                          self.fake_conf)
        self.fakedev = "/dev/gpfs0"
        self.fakefspath = "/gpfs0"
        self.fakesharepath = "/gpfs0/share-fakeid"
        self.fakeexistingshare = "existingshare"
        self.fakesnapshotpath = "/gpfs0/.snapshots/snapshot-fakesnapshotid"

        self.fake_ces_exports = """
mmcesnfslsexport:nfsexports:HEADER:version:reserved:reserved:Path:Delegations:Clients:Access_Type:Protocols:Transports:Squash:Anonymous_uid:Anonymous_gid:SecType:PrivilegedPort:DefaultDelegations:Manage_Gids:NFS_Commit:
mmcesnfslsexport:nfsexports:0:1:::/gpfs0/share-fakeid:none:44.3.2.11:RW:3,4:TCP:ROOT_SQUASH:-2:-2:SYS:FALSE:none:FALSE:FALSE:
mmcesnfslsexport:nfsexports:0:1:::/gpfs0/share-fakeid:none:1:2:3:4:5:6:7:8:RW:3,4:TCP:ROOT_SQUASH:-2:-2:SYS:FALSE:none:FALSE:FALSE:
mmcesnfslsexport:nfsexports:0:1:::/gpfs0/share-fakeid:none:10.0.0.1:RW:3,4:TCP:ROOT_SQUASH:-2:-2:SYS:FALSE:none:FALSE:FALSE:

        """
        self.fake_ces_exports_not_found = """

mmcesnfslsexport:nfsexports:HEADER:version:reserved:reserved:Path:Delegations:Clients:Access_Type:Protocols:Transports:Squash:Anonymous_uid:Anonymous_gid:SecType:PrivilegedPort:DefaultDelegations:Manage_Gids:NFS_Commit:

        """

        self.mock_object(gpfs.os.path, 'exists', mock.Mock(return_value=True))
        self._driver._helpers = {
            'CES': self._helper_fake
        }
        self.share = fake_share.fake_share(share_proto='NFS',
                                           host='fakehost@fakehost#GPFS')
        self.server = {
            'backend_details': {
                'ip': '1.2.3.4',
                'instance_id': 'fake'
            }
        }
        self.access = fake_share.fake_access()
        self.snapshot = fake_share.fake_snapshot()
        self.local_ip = "192.11.22.1"
        self.remote_ip = "192.11.22.2"
        self.remote_ip2 = "2.2.2.2"
        gpfs_nfs_server_list = [self.remote_ip, self.local_ip, self.remote_ip2,
                                "fake_location"]
        self._knfs_helper.configuration.gpfs_nfs_server_list = (
            gpfs_nfs_server_list)
        self._ces_helper.configuration.gpfs_nfs_server_list = (
            gpfs_nfs_server_list)
        self._ces_helper.configuration.ganesha_config_path = (
            "fake_ganesha_config_path")
        self.sshlogin = "fake_login"
        self.sshkey = "fake_sshkey"
        self.gservice = "fake_ganesha_service"
        self._ces_helper.configuration.gpfs_ssh_login = self.sshlogin
        self._ces_helper.configuration.gpfs_ssh_private_key = self.sshkey
        self._ces_helper.configuration.ganesha_service_name = self.gservice
        self.mock_object(socket, 'gethostname',
                         mock.Mock(return_value="testserver"))
        self.mock_object(socket, 'gethostbyname_ex', mock.Mock(
            return_value=('localhost',
                          ['localhost.localdomain', 'testserver'],
                          ['127.0.0.1', self.local_ip])
        ))

    def test__run_ssh(self):
        cmd_list = ['fake', 'cmd']
        expected_cmd = 'fake cmd'
        ssh_pool = mock.Mock()
        ssh = mock.Mock()
        self.mock_object(utils, 'SSHPool', mock.Mock(return_value=ssh_pool))
        ssh_pool.item = mock.Mock(return_value=ssh)
        setattr(ssh, '__enter__', mock.Mock())
        setattr(ssh, '__exit__', mock.Mock())
        self.mock_object(self._driver, '_gpfs_ssh_execute')
        self._driver._run_ssh(self.local_ip, cmd_list)

        self._driver._gpfs_ssh_execute.assert_called_once_with(
            mock.ANY, expected_cmd, check_exit_code=True,
            ignore_exit_code=None)

    def test__run_ssh_exception(self):
        cmd_list = ['fake', 'cmd']
        ssh_pool = mock.Mock()
        ssh = mock.Mock()
        self.mock_object(utils, 'SSHPool', mock.Mock(return_value=ssh_pool))
        ssh_pool.item = mock.Mock(return_value=ssh)
        self.mock_object(self._driver, '_gpfs_ssh_execute')
        self.assertRaises(exception.GPFSException,
                          self._driver._run_ssh,
                          self.local_ip, cmd_list)

    def test__gpfs_ssh_execute(self):
        cmd = 'fake cmd'
        expected_out = 'cmd successful'
        expected_err = 'cmd error'
        ssh = mock.Mock()
        stdin_stream = mock.Mock()
        stdout_stream = mock.Mock()
        stderr_stream = mock.Mock()
        ssh.exec_command = mock.Mock(return_value=(stdin_stream,
                                                   stdout_stream,
                                                   stderr_stream))
        stdout_stream.channel.recv_exit_status = mock.Mock(return_value=-1)
        stdout_stream.read = mock.Mock(return_value=expected_out)
        stderr_stream.read = mock.Mock(return_value=expected_err)
        stdin_stream.close = mock.Mock()
        actual_out, actual_err = self._driver._gpfs_ssh_execute(ssh, cmd)

        self.assertEqual(actual_out, expected_out)
        self.assertEqual(actual_err, expected_err)

    def test__gpfs_ssh_execute_exception(self):
        cmd = 'fake cmd'
        ssh = mock.Mock()
        stdin_stream = mock.Mock()
        stdout_stream = mock.Mock()
        stderr_stream = mock.Mock()
        ssh.exec_command = mock.Mock(return_value=(stdin_stream,
                                                   stdout_stream,
                                                   stderr_stream))
        stdout_stream.channel.recv_exit_status = mock.Mock(return_value=1)
        stdout_stream.read = mock.Mock()
        stderr_stream.read = mock.Mock()
        stdin_stream.close = mock.Mock()
        self.assertRaises(exception.ProcessExecutionError,
                          self._driver._gpfs_ssh_execute,
                          ssh, cmd)

    def test_get_share_stats_refresh_false(self):
        self._driver._stats = {'fake_key': 'fake_value'}
        result = self._driver.get_share_stats(False)
        self.assertEqual(self._driver._stats, result)

    def test_get_share_stats_refresh_true(self):
        self.mock_object(
            self._driver, '_get_available_capacity',
            mock.Mock(return_value=(11111.0, 12345.0)))
        result = self._driver.get_share_stats(True)
        expected_keys = [
            'qos', 'driver_version', 'share_backend_name',
            'free_capacity_gb', 'total_capacity_gb',
            'driver_handles_share_servers',
            'reserved_percentage', 'vendor_name', 'storage_protocol',
        ]
        for key in expected_keys:
            self.assertIn(key, result)
        self.assertFalse(result['driver_handles_share_servers'])
        self.assertEqual('IBM', result['vendor_name'])
        self._driver._get_available_capacity.assert_called_once_with(
            self._driver.configuration.gpfs_mount_point_base)

    def test_do_setup(self):
        self.mock_object(self._driver, '_setup_helpers')
        self._driver.do_setup(self._context)
        self.assertEqual(self._driver._gpfs_execute,
                         self._driver._gpfs_remote_execute)
        self._driver._setup_helpers.assert_called_once_with()

    def test_do_setup_gpfs_local_execute(self):
        self.mock_object(self._driver, '_setup_helpers')
        self._driver.configuration.is_gpfs_node = True
        self._driver.do_setup(self._context)
        self.assertEqual(self._driver._gpfs_execute,
                         self._driver._gpfs_local_execute)
        self._driver._setup_helpers.assert_called_once_with()

    def test_setup_helpers(self):
        self._driver._helpers = {}
        CONF.set_default('gpfs_share_helpers', ['CES=fakenfs'])
        self.mock_object(gpfs.importutils, 'import_class',
                         mock.Mock(return_value=self._helper_fake))
        self._driver._setup_helpers()
        gpfs.importutils.import_class.assert_has_calls(
            [mock.call('fakenfs')]
        )
        self.assertEqual(len(self._driver._helpers), 1)

    @ddt.data(fake_share.fake_share(),
              fake_share.fake_share(share_proto='NFSBOGUS'))
    def test__get_helper_with_wrong_proto(self, share):
        self.assertRaises(exception.InvalidShare,
                          self._driver._get_helper, share)

    def test__local_path(self):
        sharename = 'fakesharename'
        self._driver.configuration.gpfs_mount_point_base = (
            self.fakefspath)
        local_path = self._driver._local_path(sharename)
        self.assertEqual(self.fakefspath + '/' + sharename,
                         local_path)

    def test__get_share_path(self):
        self._driver.configuration.gpfs_mount_point_base = (
            self.fakefspath)
        share_path = self._driver._get_share_path(self.share)
        self.assertEqual(self.fakefspath + '/' + self.share['name'],
                         share_path)

    def test__get_snapshot_path(self):
        self._driver.configuration.gpfs_mount_point_base = (
            self.fakefspath)
        snapshot_path = self._driver._get_snapshot_path(self.snapshot)
        self.assertEqual(self.fakefspath + '/' + self.snapshot['share_name'] +
                         '/.snapshots/' + self.snapshot['name'],
                         snapshot_path)

    def test_check_for_setup_error_for_gpfs_state(self):
        self.mock_object(self._driver, '_check_gpfs_state',
                         mock.Mock(return_value=False))
        self.assertRaises(exception.GPFSException,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_for_export_ip(self):
        self.mock_object(self._driver, '_check_gpfs_state',
                         mock.Mock(return_value=True))

        self._driver.configuration.gpfs_share_export_ip = None
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_for_gpfs_mount_point_base(self):
        self.mock_object(self._driver, '_check_gpfs_state',
                         mock.Mock(return_value=True))
        self._driver.configuration.gpfs_share_export_ip = self.local_ip
        self._driver.configuration.gpfs_mount_point_base = 'test'
        self.assertRaises(exception.GPFSException,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_for_directory_check(self):
        self.mock_object(self._driver, '_check_gpfs_state',
                         mock.Mock(return_value=True))
        self._driver.configuration.gpfs_share_export_ip = self.local_ip
        self._driver.configuration.gpfs_mount_point_base = self.fakefspath
        self.mock_object(self._driver, '_is_dir',
                         mock.Mock(return_value=False))
        self.assertRaises(exception.GPFSException,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_for_gpfs_path_check(self):
        self.mock_object(self._driver, '_check_gpfs_state',
                         mock.Mock(return_value=True))
        self._driver.configuration.gpfs_share_export_ip = self.local_ip
        self._driver.configuration.gpfs_mount_point_base = self.fakefspath
        self.mock_object(self._driver, '_is_dir',
                         mock.Mock(return_value=True))
        self.mock_object(self._driver, '_is_gpfs_path',
                         mock.Mock(return_value=False))
        self.assertRaises(exception.GPFSException,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_for_nfs_server_type(self):
        self.mock_object(self._driver, '_check_gpfs_state',
                         mock.Mock(return_value=True))
        self._driver.configuration.gpfs_share_export_ip = self.local_ip
        self._driver.configuration.gpfs_mount_point_base = self.fakefspath
        self.mock_object(self._driver, '_is_dir',
                         mock.Mock(return_value=True))
        self.mock_object(self._driver, '_is_gpfs_path',
                         mock.Mock(return_value=True))
        self._driver.configuration.gpfs_nfs_server_type = 'test'
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test_check_for_setup_error_for_nfs_server_list(self):
        self.mock_object(self._driver, '_check_gpfs_state',
                         mock.Mock(return_value=True))
        self._driver.configuration.gpfs_share_export_ip = self.local_ip
        self._driver.configuration.gpfs_mount_point_base = self.fakefspath
        self.mock_object(self._driver, '_is_dir',
                         mock.Mock(return_value=True))
        self.mock_object(self._driver, '_is_gpfs_path',
                         mock.Mock(return_value=True))
        self._driver.configuration.gpfs_nfs_server_type = 'KNFS'
        self._driver.configuration.gpfs_nfs_server_list = None
        self.assertRaises(exception.InvalidParameterValue,
                          self._driver.check_for_setup_error)

    def test__get_available_capacity(self):
        path = self.fakefspath
        mock_out = "Filesystem 1-blocks Used Available Capacity Mounted on\n\
                    /dev/gpfs0 100 30 70 30% /gpfs0"
        self.mock_object(self._driver, '_gpfs_execute',
                         mock.Mock(return_value=(mock_out, '')))
        available, size = self._driver._get_available_capacity(path)
        self.assertEqual(70, available)
        self.assertEqual(100, size)

    def test_create_share(self):
        self._helper_fake.create_export.return_value = 'fakelocation'
        methods = ('_create_share', '_get_share_path')
        for method in methods:
            self.mock_object(self._driver, method)
        result = self._driver.create_share(self._context, self.share,
                                           share_server=self.server)
        self._driver._create_share.assert_called_once_with(self.share)
        self._driver._get_share_path.assert_called_once_with(self.share)

        self.assertEqual(result, 'fakelocation')

    def test_create_share_from_snapshot(self):
        self._helper_fake.create_export.return_value = 'fakelocation'
        self._driver._get_share_path = mock.Mock(return_value=self.
                                                 fakesharepath)
        self._driver._create_share_from_snapshot = mock.Mock()
        result = self._driver.create_share_from_snapshot(self._context,
                                                         self.share,
                                                         self.snapshot,
                                                         share_server=None)
        self._driver._get_share_path.assert_called_once_with(self.share)
        self._driver._create_share_from_snapshot.assert_called_once_with(
            self.share, self.snapshot,
            self.fakesharepath
        )
        self.assertEqual(result, 'fakelocation')

    def test_create_snapshot(self):
        self._driver._create_share_snapshot = mock.Mock()
        self._driver.create_snapshot(self._context, self.snapshot,
                                     share_server=None)
        self._driver._create_share_snapshot.assert_called_once_with(
            self.snapshot
        )

    def test_delete_share(self):
        self._driver._get_share_path = mock.Mock(
            return_value=self.fakesharepath
        )
        self._driver._delete_share = mock.Mock()

        self._driver.delete_share(self._context, self.share,
                                  share_server=None)

        self._driver._get_share_path.assert_called_once_with(self.share)
        self._driver._delete_share.assert_called_once_with(self.share)
        self._helper_fake.remove_export.assert_called_once_with(
            self.fakesharepath, self.share
        )

    def test_delete_snapshot(self):
        self._driver._delete_share_snapshot = mock.Mock()
        self._driver.delete_snapshot(self._context, self.snapshot,
                                     share_server=None)
        self._driver._delete_share_snapshot.assert_called_once_with(
            self.snapshot
        )

    def test__delete_share_snapshot(self):
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._gpfs_execute = mock.Mock(return_value=0)
        self._driver._delete_share_snapshot(self.snapshot)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmdelsnapshot', self.fakedev,
            self.snapshot['name'], '-j', self.snapshot['share_name']
        )
        self._driver._get_gpfs_device.assert_called_once_with()

    def test__delete_share_snapshot_exception(self):
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._delete_share_snapshot, self.snapshot)
        self._driver._get_gpfs_device.assert_called_once_with()
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmdelsnapshot', self.fakedev,
            self.snapshot['name'], '-j', self.snapshot['share_name']
        )

    def test_extend_share(self):
        self._driver._extend_share = mock.Mock()
        self._driver.extend_share(self.share, 10)
        self._driver._extend_share.assert_called_once_with(self.share, 10)

    def test__extend_share(self):
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._gpfs_execute = mock.Mock(return_value=True)
        self._driver._extend_share(self.share, 10)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmsetquota', self.fakedev + ':' +
            self.share['name'], '--block', '0:10G')
        self._driver._get_gpfs_device.assert_called_once_with()

    def test__extend_share_exception(self):
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._extend_share, self.share, 10)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmsetquota', self.fakedev + ':' +
            self.share['name'], '--block', '0:10G')
        self._driver._get_gpfs_device.assert_called_once_with()

    def test_update_access_allow(self):
        """Test allow_access functionality via update_access."""
        self._driver._get_share_path = mock.Mock(
            return_value=self.fakesharepath
        )
        self._helper_fake.allow_access = mock.Mock()

        self._driver.update_access(self._context,
                                   self.share,
                                   ["ignored"],
                                   [self.access],
                                   [],
                                   share_server=None)

        self._helper_fake.allow_access.assert_called_once_with(
            self.fakesharepath, self.share, self.access)
        self.assertFalse(self._helper_fake.resync_access.called)
        self._driver._get_share_path.assert_called_once_with(self.share)

    def test_update_access_deny(self):
        """Test deny_access functionality via update_access."""
        self._driver._get_share_path = mock.Mock(return_value=self.
                                                 fakesharepath)
        self._helper_fake.deny_access = mock.Mock()

        self._driver.update_access(self._context,
                                   self.share,
                                   ["ignored"],
                                   [],
                                   [self.access],
                                   share_server=None)

        self._helper_fake.deny_access.assert_called_once_with(
            self.fakesharepath, self.share, self.access)
        self.assertFalse(self._helper_fake.resync_access.called)
        self._driver._get_share_path.assert_called_once_with(self.share)

    def test_update_access_both(self):
        """Test update_access with allow and deny lists."""
        self._driver._get_share_path = mock.Mock(return_value=self.
                                                 fakesharepath)
        self._helper_fake.deny_access = mock.Mock()
        self._helper_fake.allow_access = mock.Mock()
        self._helper_fake.resync_access = mock.Mock()

        access_1 = fake_share.fake_access(access_to="1.1.1.1")
        access_2 = fake_share.fake_access(access_to="2.2.2.2")
        self._driver.update_access(self._context,
                                   self.share,
                                   ["ignore"],
                                   [access_1],
                                   [access_2],
                                   share_server=None)

        self.assertFalse(self._helper_fake.resync_access.called)
        self._helper_fake.allow_access.assert_called_once_with(
            self.fakesharepath, self.share, access_1)
        self._helper_fake.deny_access.assert_called_once_with(
            self.fakesharepath, self.share, access_2)
        self._driver._get_share_path.assert_called_once_with(self.share)

    def test_update_access_resync(self):
        """Test recovery mode update_access."""
        self._driver._get_share_path = mock.Mock(return_value=self.
                                                 fakesharepath)
        self._helper_fake.deny_access = mock.Mock()
        self._helper_fake.allow_access = mock.Mock()
        self._helper_fake.resync_access = mock.Mock()

        access_1 = fake_share.fake_access(access_to="1.1.1.1")
        access_2 = fake_share.fake_access(access_to="2.2.2.2")
        self._driver.update_access(self._context,
                                   self.share,
                                   [access_1, access_2],
                                   [],
                                   [],
                                   share_server=None)

        self._helper_fake.resync_access.assert_called_once_with(
            self.fakesharepath, self.share, [access_1, access_2])
        self.assertFalse(self._helper_fake.allow_access.called)
        self.assertFalse(self._helper_fake.allow_access.called)
        self._driver._get_share_path.assert_called_once_with(self.share)

    def test__check_gpfs_state_active(self):
        fakeout = "mmgetstate::state:\nmmgetstate::active:"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))
        result = self._driver._check_gpfs_state()
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmgetstate', '-Y')
        self.assertEqual(result, True)

    def test__check_gpfs_state_down(self):
        fakeout = "mmgetstate::state:\nmmgetstate::down:"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))
        result = self._driver._check_gpfs_state()
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmgetstate', '-Y')
        self.assertEqual(result, False)

    def test__check_gpfs_state_wrong_output_exception(self):
        fakeout = "mmgetstate fake out"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))
        self.assertRaises(exception.GPFSException,
                          self._driver._check_gpfs_state)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmgetstate', '-Y')

    def test__check_gpfs_state_exception(self):
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._check_gpfs_state)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmgetstate', '-Y')

    def test__is_dir_success(self):
        fakeoutput = "directory"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeoutput, ''))
        result = self._driver._is_dir(self.fakefspath)
        self._driver._gpfs_execute.assert_called_once_with(
            'stat', '--format=%F', self.fakefspath, run_as_root=False
        )
        self.assertEqual(result, True)

    def test__is_dir_failure(self):
        fakeoutput = "regular file"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeoutput, ''))
        result = self._driver._is_dir(self.fakefspath)
        self._driver._gpfs_execute.assert_called_once_with(
            'stat', '--format=%F', self.fakefspath, run_as_root=False
        )
        self.assertEqual(result, False)

    def test__is_dir_exception(self):
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._is_dir, self.fakefspath)
        self._driver._gpfs_execute.assert_called_once_with(
            'stat', '--format=%F', self.fakefspath, run_as_root=False
        )

    def test__is_gpfs_path_ok(self):
        self._driver._gpfs_execute = mock.Mock(return_value=0)
        result = self._driver._is_gpfs_path(self.fakefspath)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmlsattr', self.fakefspath)
        self.assertEqual(result, True)

    def test__is_gpfs_path_exception(self):
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._is_gpfs_path,
                          self.fakefspath)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmlsattr', self.fakefspath)

    def test__get_gpfs_device(self):
        fakeout = "Filesystem\n" + self.fakedev
        orig_val = self._driver.configuration.gpfs_mount_point_base
        self._driver.configuration.gpfs_mount_point_base = self.fakefspath
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))
        result = self._driver._get_gpfs_device()
        self._driver._gpfs_execute.assert_called_once_with('df',
                                                           self.fakefspath)
        self.assertEqual(result, self.fakedev)
        self._driver.configuration.gpfs_mount_point_base = orig_val

    def test__get_gpfs_device_exception(self):
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.GPFSException,
                          self._driver._get_gpfs_device)

    def test__create_share(self):
        sizestr = '%sG' % self.share['size']
        self._driver._gpfs_execute = mock.Mock(return_value=True)
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._create_share(self.share)
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmcrfileset', self.fakedev, self.share['name'],
            '--inode-space', 'new')
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmlinkfileset', self.fakedev, self.share['name'],
            '-J', self.fakesharepath)
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmsetquota', self.fakedev + ':' +
            self.share['name'], '--block', '0:' + sizestr)
        self._driver._gpfs_execute.assert_any_call(
            'chmod', '777', self.fakesharepath)

        self._driver._local_path.assert_called_once_with(self.share['name'])
        self._driver._get_gpfs_device.assert_called_once_with()

    def test__create_share_exception(self):
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._create_share, self.share)
        self._driver._get_gpfs_device.assert_called_once_with()
        self._driver._local_path.assert_called_once_with(self.share['name'])
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmcrfileset', self.fakedev, self.share['name'],
            '--inode-space', 'new')

    def test__delete_share(self):
        self._driver._gpfs_execute = mock.Mock(return_value=True)
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._delete_share(self.share)
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
            self.share['name'], '-f', ignore_exit_code=[2])
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmdelfileset', self.fakedev, self.share['name'],
            '-f', ignore_exit_code=[2])
        self._driver._get_gpfs_device.assert_called_once_with()

    def test__delete_share_exception(self):
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._delete_share, self.share)
        self._driver._get_gpfs_device.assert_called_once_with()
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
            self.share['name'], '-f', ignore_exit_code=[2])

    def test__create_share_snapshot(self):
        self._driver._gpfs_execute = mock.Mock(return_value=True)
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._create_share_snapshot(self.snapshot)
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmcrsnapshot', self.fakedev,
            self.snapshot['name'], '-j', self.snapshot['share_name']
        )
        self._driver._get_gpfs_device.assert_called_once_with()

    def test__create_share_snapshot_exception(self):
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._create_share_snapshot, self.snapshot)
        self._driver._get_gpfs_device.assert_called_once_with()
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmcrsnapshot', self.fakedev,
            self.snapshot['name'], '-j', self.snapshot['share_name']
        )

    def test__create_share_from_snapshot(self):
        self._driver._gpfs_execute = mock.Mock(return_value=True)
        self._driver._create_share = mock.Mock(return_value=True)
        self._driver._get_snapshot_path = mock.Mock(return_value=self.
                                                    fakesnapshotpath)
        self._driver._create_share_from_snapshot(self.share, self.snapshot,
                                                 self.fakesharepath)
        self._driver._gpfs_execute.assert_called_once_with(
            'rsync', '-rp', self.fakesnapshotpath + '/', self.fakesharepath
        )
        self._driver._create_share.assert_called_once_with(self.share)
        self._driver._get_snapshot_path.assert_called_once_with(self.snapshot)

    def test__create_share_from_snapshot_exception(self):
        self._driver._create_share = mock.Mock(return_value=True)
        self._driver._get_snapshot_path = mock.Mock(return_value=self.
                                                    fakesnapshotpath)
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        self.assertRaises(exception.GPFSException,
                          self._driver._create_share_from_snapshot,
                          self.share, self.snapshot, self.fakesharepath)
        self._driver._create_share.assert_called_once_with(self.share)
        self._driver._get_snapshot_path.assert_called_once_with(self.snapshot)
        self._driver._gpfs_execute.assert_called_once_with(
            'rsync', '-rp', self.fakesnapshotpath + '/', self.fakesharepath
        )

    @ddt.data("mmlsfileset::allocInodes:\nmmlsfileset::100096:",
              "mmlsfileset::allocInodes:\nmmlsfileset::0:")
    def test__is_share_valid_with_quota(self, fakeout):
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))

        result = self._driver._is_share_valid(self.fakedev, self.fakesharepath)

        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmlsfileset', self.fakedev, '-J',
            self.fakesharepath, '-L', '-Y')
        if fakeout == "mmlsfileset::allocInodes:\nmmlsfileset::100096:":
            self.assertTrue(result)
        else:
            self.assertFalse(result)

    def test__is_share_valid_exception(self):
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)

        self.assertRaises(exception.ManageInvalidShare,
                          self._driver._is_share_valid, self.fakedev,
                          self.fakesharepath)

        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmlsfileset', self.fakedev, '-J',
            self.fakesharepath, '-L', '-Y')

    def test__is_share_valid_no_share_exist_exception(self):
        fakeout = "mmlsfileset::allocInodes:"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))

        self.assertRaises(exception.GPFSException,
                          self._driver._is_share_valid, self.fakedev,
                          self.fakesharepath)

        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmlsfileset', self.fakedev, '-J',
            self.fakesharepath, '-L', '-Y')

    def test__get_share_name(self):
        fakeout = "mmlsfileset::filesetName:\nmmlsfileset::existingshare:"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))

        result = self._driver._get_share_name(self.fakedev, self.fakesharepath)

        self.assertEqual('existingshare', result)

    def test__get_share_name_exception(self):
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)

        self.assertRaises(exception.ManageInvalidShare,
                          self._driver._get_share_name, self.fakedev,
                          self.fakesharepath)

        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmlsfileset', self.fakedev, '-J',
            self.fakesharepath, '-L', '-Y')

    def test__get_share_name_no_share_exist_exception(self):
        fakeout = "mmlsfileset::filesetName:"
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))

        self.assertRaises(exception.GPFSException,
                          self._driver._get_share_name, self.fakedev,
                          self.fakesharepath)

        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmlsfileset', self.fakedev, '-J',
            self.fakesharepath, '-L', '-Y')

    @ddt.data("mmlsquota::blockLimit:\nmmlsquota::1048577",
              "mmlsquota::blockLimit:\nmmlsquota::1048576",
              "mmlsquota::blockLimit:\nmmlsquota::0")
    def test__manage_existing(self, fakeout):
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))
        self._helper_fake.create_export.return_value = 'fakelocation'
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)

        actual_size, actual_path = self._driver._manage_existing(
            self.fakedev, self.share, self.fakeexistingshare)

        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
            self.fakeexistingshare, '-f')
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmchfileset', self.fakedev,
            self.fakeexistingshare, '-j', self.share['name'])
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmlinkfileset', self.fakedev, self.share['name'],
            '-J', self.fakesharepath)
        self._driver._gpfs_execute.assert_any_call(
            'chmod', '777', self.fakesharepath)
        if fakeout == "mmlsquota::blockLimit:\nmmlsquota::1048577":
            self._driver._gpfs_execute.assert_called_with(
                self.GPFS_PATH + 'mmsetquota', self.fakedev + ':' +
                self.share['name'], '--block', '0:2G')
            self.assertEqual(2, actual_size)
            self.assertEqual('fakelocation', actual_path)
        elif fakeout == "mmlsquota::blockLimit:\nmmlsquota::0":
            self._driver._gpfs_execute.assert_called_with(
                self.GPFS_PATH + 'mmsetquota', self.fakedev + ':' +
                self.share['name'], '--block', '0:1G')
            self.assertEqual(1, actual_size)
            self.assertEqual('fakelocation', actual_path)
        else:
            self.assertEqual(1, actual_size)
            self.assertEqual('fakelocation', actual_path)

    def test__manage_existing_fileset_unlink_exception(self):
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self._driver._gpfs_execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_called_once_with(self.share['name'])
        self._driver._gpfs_execute.assert_called_once_with(
            self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
            self.fakeexistingshare, '-f')

    def test__manage_existing_fileset_creation_exception(self):
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self.mock_object(self._driver, '_gpfs_execute', mock.Mock(
            side_effect=['', exception.ProcessExecutionError]))

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_any_call(self.share['name'])
        self._driver._gpfs_execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
                      self.fakeexistingshare, '-f'),
            mock.call(self.GPFS_PATH + 'mmchfileset', self.fakedev,
                      self.fakeexistingshare, '-j', self.share['name'])])

    def test__manage_existing_fileset_relink_exception(self):
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self.mock_object(self._driver, '_gpfs_execute', mock.Mock(
            side_effect=['', '', exception.ProcessExecutionError]))

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_any_call(self.share['name'])
        self._driver._gpfs_execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
                      self.fakeexistingshare, '-f'),
            mock.call(self.GPFS_PATH + 'mmchfileset', self.fakedev,
                      self.fakeexistingshare, '-j', self.share['name']),
            mock.call(self.GPFS_PATH + 'mmlinkfileset', self.fakedev,
                      self.share['name'], '-J', self.fakesharepath)])

    def test__manage_existing_permission_change_exception(self):
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self.mock_object(self._driver, '_gpfs_execute', mock.Mock(
            side_effect=['', '', '', exception.ProcessExecutionError]))

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_any_call(self.share['name'])
        self._driver._gpfs_execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
                      self.fakeexistingshare, '-f'),
            mock.call(self.GPFS_PATH + 'mmchfileset', self.fakedev,
                      self.fakeexistingshare, '-j', self.share['name']),
            mock.call(self.GPFS_PATH + 'mmlinkfileset', self.fakedev,
                      self.share['name'], '-J', self.fakesharepath),
            mock.call('chmod', '777', self.fakesharepath)])

    def test__manage_existing_checking_quota_of_fileset_exception(self):
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self.mock_object(self._driver, '_gpfs_execute', mock.Mock(
            side_effect=['', '', '', '', exception.ProcessExecutionError]))

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_any_call(self.share['name'])
        self._driver._gpfs_execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
                      self.fakeexistingshare, '-f'),
            mock.call(self.GPFS_PATH + 'mmchfileset', self.fakedev,
                      self.fakeexistingshare, '-j', self.share['name']),
            mock.call(self.GPFS_PATH + 'mmlinkfileset', self.fakedev,
                      self.share['name'], '-J', self.fakesharepath),
            mock.call('chmod', '777', self.fakesharepath),
            mock.call(self.GPFS_PATH + 'mmlsquota', '-j', self.share['name'],
                      '-Y', self.fakedev)])

    def test__manage_existing_unable_to_get_quota_of_fileset_exception(self):
        fakeout = "mmlsquota::blockLimit:"
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self._driver._gpfs_execute = mock.Mock(return_value=(fakeout, ''))

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_any_call(self.share['name'])
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
            self.fakeexistingshare, '-f')
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmchfileset', self.fakedev,
            self.fakeexistingshare, '-j', self.share['name'])
        self._driver._gpfs_execute.assert_any_call(
            self.GPFS_PATH + 'mmlinkfileset', self.fakedev,
            self.share['name'], '-J', self.fakesharepath)
        self._driver._gpfs_execute.assert_any_call(
            'chmod', '777', self.fakesharepath)
        self._driver._gpfs_execute.assert_called_with(
            self.GPFS_PATH + 'mmlsquota', '-j', self.share['name'],
            '-Y', self.fakedev)

    def test__manage_existing_set_quota_of_fileset_less_than_1G_exception(
            self):
        sizestr = '1G'
        mock_out = "mmlsquota::blockLimit:\nmmlsquota::0:", None
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self.mock_object(self._driver, '_gpfs_execute', mock.Mock(
            side_effect=['', '', '', '', mock_out,
                         exception.ProcessExecutionError]))

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_any_call(self.share['name'])
        self._driver._gpfs_execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
                      self.fakeexistingshare, '-f'),
            mock.call(self.GPFS_PATH + 'mmchfileset', self.fakedev,
                      self.fakeexistingshare, '-j', self.share['name']),
            mock.call(self.GPFS_PATH + 'mmlinkfileset', self.fakedev,
                      self.share['name'], '-J', self.fakesharepath),
            mock.call('chmod', '777', self.fakesharepath),
            mock.call(self.GPFS_PATH + 'mmlsquota', '-j', self.share['name'],
                      '-Y', self.fakedev),
            mock.call(self.GPFS_PATH + 'mmsetquota', self.fakedev + ':' +
                      self.share['name'], '--block', '0:' + sizestr)])

    def test__manage_existing_set_quota_of_fileset_grater_than_1G_exception(
            self):
        sizestr = '2G'
        mock_out = "mmlsquota::blockLimit:\nmmlsquota::1048577:", None
        self._driver._local_path = mock.Mock(return_value=self.fakesharepath)
        self.mock_object(self._driver, '_gpfs_execute', mock.Mock(
            side_effect=['', '', '', '', mock_out,
                         exception.ProcessExecutionError]))

        self.assertRaises(exception.GPFSException,
                          self._driver._manage_existing, self.fakedev,
                          self.share, self.fakeexistingshare)

        self._driver._local_path.assert_any_call(self.share['name'])
        self._driver._gpfs_execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmunlinkfileset', self.fakedev,
                      self.fakeexistingshare, '-f'),
            mock.call(self.GPFS_PATH + 'mmchfileset', self.fakedev,
                      self.fakeexistingshare, '-j', self.share['name']),
            mock.call(self.GPFS_PATH + 'mmlinkfileset', self.fakedev,
                      self.share['name'], '-J', self.fakesharepath),
            mock.call('chmod', '777', self.fakesharepath),
            mock.call(self.GPFS_PATH + 'mmlsquota', '-j', self.share['name'],
                      '-Y', self.fakedev),
            mock.call(self.GPFS_PATH + 'mmsetquota', self.fakedev + ':' +
                      self.share['name'], '--block', '0:' + sizestr)])

    def test_manage_existing(self):
        self._driver._manage_existing = mock.Mock(return_value=('1',
                                                  'fakelocation'))
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._is_share_valid = mock.Mock(return_value=True)
        self._driver._get_share_name = mock.Mock(return_value=self.
                                                 fakeexistingshare)
        self._helper_fake._has_client_access = mock.Mock(return_value=[])

        result = self._driver.manage_existing(self.share, {})

        self.assertEqual('1', result['size'])
        self.assertEqual('fakelocation', result['export_locations'])

    def test_manage_existing_incorrect_path_exception(self):
        share = fake_share.fake_share(export_location="wrong_ip::wrong_path")

        self.assertRaises(exception.ShareBackendException,
                          self._driver.manage_existing, share, {})

    def test_manage_existing_incorrect_ip_exception(self):
        share = fake_share.fake_share(export_location="wrong_ip:wrong_path")

        self.assertRaises(exception.ShareBackendException,
                          self._driver.manage_existing, share, {})

    def test__manage_existing_invalid_export_exception(self):
        share = fake_share.fake_share(export_location="wrong_ip/wrong_path")

        self.assertRaises(exception.ShareBackendException,
                          self._driver.manage_existing, share, {})

    @ddt.data(True, False)
    def test_manage_existing_invalid_share_exception(self, valid_share):
        self._driver._get_gpfs_device = mock.Mock(return_value=self.fakedev)
        self._driver._is_share_valid = mock.Mock(return_value=valid_share)
        if valid_share:
            self._driver._get_share_name = mock.Mock(return_value=self.
                                                     fakeexistingshare)
            self._helper_fake._has_client_access = mock.Mock()
        else:
            self.assertFalse(self._helper_fake._has_client_access.called)

        self.assertRaises(exception.ManageInvalidShare,
                          self._driver.manage_existing, self.share, {})

    def test__gpfs_local_execute(self):
        self.mock_object(utils, 'execute', mock.Mock(return_value=True))
        cmd = "testcmd"
        self._driver._gpfs_local_execute(cmd, ignore_exit_code=[2])
        utils.execute.assert_called_once_with(cmd, run_as_root=True,
                                              check_exit_code=[2, 0])

    def test__gpfs_remote_execute(self):
        self._driver._run_ssh = mock.Mock(return_value=True)
        cmd = "testcmd"
        orig_value = self._driver.configuration.gpfs_share_export_ip
        self._driver.configuration.gpfs_share_export_ip = self.local_ip
        self._driver._gpfs_remote_execute(cmd, check_exit_code=True)
        self._driver._run_ssh.assert_called_once_with(
            self.local_ip, tuple([cmd]), None, True
        )
        self._driver.configuration.gpfs_share_export_ip = orig_value

    def test_knfs_resync_access(self):
        self._knfs_helper.allow_access = mock.Mock()
        path = self.fakesharepath
        to_remove = '3.3.3.3'
        fake_exportfs_before = ('%(path)s\n\t\t%(ip)s\n'
                                '/other/path\n\t\t4.4.4.4\n' %
                                {'path': path, 'ip': to_remove})
        fake_exportfs_after = '/other/path\n\t\t4.4.4.4\n'
        self._knfs_helper._execute = mock.Mock(
            return_value=(fake_exportfs_before, ''))
        self._knfs_helper._publish_access = mock.Mock(
            side_effect=[[(fake_exportfs_before, '')],
                         [(fake_exportfs_after, '')]])

        access_1 = fake_share.fake_access(access_to="1.1.1.1")
        access_2 = fake_share.fake_access(access_to="2.2.2.2")
        self._knfs_helper.resync_access(path, self.share, [access_1, access_2])

        self._knfs_helper.allow_access.assert_has_calls([
            mock.call(path, self.share, access_1, error_on_exists=False),
            mock.call(path, self.share, access_2, error_on_exists=False)])
        self._knfs_helper._execute.assert_called_once_with(
            'exportfs', run_as_root=True)
        self._knfs_helper._publish_access.assert_has_calls([
            mock.call('exportfs', '-u',
                      '%(ip)s:%(path)s' % {'ip': to_remove, 'path': path},
                      check_exit_code=[0, 1]),
            mock.call('exportfs')])

    @ddt.data('rw', 'ro')
    def test_knfs_get_export_options(self, access_level):
        mock_out = {"knfs:export_options": "no_root_squash"}
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value=mock_out))
        access = fake_share.fake_access(access_level=access_level)
        out = self._knfs_helper.get_export_options(self.share, access, 'KNFS')
        self.assertEqual("no_root_squash,%s" % access_level, out)

    def test_knfs_get_export_options_default(self):
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value={}))
        access = self.access
        out = self._knfs_helper.get_export_options(self.share, access, 'KNFS')
        self.assertEqual("rw", out)

    def test_knfs_get_export_options_invalid_option_ro(self):
        mock_out = {"knfs:export_options": "ro"}
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value=mock_out))
        access = self.access
        share = fake_share.fake_share(share_type="fake_share_type")
        self.assertRaises(exception.InvalidInput,
                          self._knfs_helper.get_export_options,
                          share, access, 'KNFS')

    def test_knfs_get_export_options_invalid_option_rw(self):
        mock_out = {"knfs:export_options": "rw"}
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value=mock_out))
        access = self.access
        share = fake_share.fake_share(share_type="fake_share_type")
        self.assertRaises(exception.InvalidInput,
                          self._knfs_helper.get_export_options,
                          share, access, 'KNFS')

    @ddt.data(("/gpfs0/share-fakeid\t10.0.0.1", None),
              ("", None),
              ("/gpfs0/share-fakeid\t10.0.0.1", "10.0.0.1"),
              ("/gpfs0/share-fakeid\t10.0.0.1", "10.0.0.2"))
    @ddt.unpack
    def test_knfs__has_client_access(self, mock_out, access_to):
        self._knfs_helper._execute = mock.Mock(return_value=[mock_out, 0])

        result = self._knfs_helper._has_client_access(self.fakesharepath,
                                                      access_to)

        self._ces_helper._execute.assert_called_once_with('exportfs',
                                                          check_exit_code=True,
                                                          run_as_root=True)
        if mock_out == "/gpfs0/share-fakeid\t10.0.0.1":
            if access_to in (None, "10.0.0.1"):
                self.assertTrue(result)
            else:
                self.assertFalse(result)
        else:
            self.assertFalse(result)

    def test_knfs_allow_access(self):
        self._knfs_helper._execute = mock.Mock(
            return_value=['/fs0 <world>', 0]
        )
        self.mock_object(re, 'search', mock.Mock(return_value=None))
        export_opts = None
        self._knfs_helper.get_export_options = mock.Mock(
            return_value=export_opts
        )
        self._knfs_helper._publish_access = mock.Mock()
        access = self.access
        local_path = self.fakesharepath
        self._knfs_helper.allow_access(local_path, self.share, access)
        self._knfs_helper._execute.assert_called_once_with('exportfs',
                                                           run_as_root=True)
        self.assertTrue(re.search.called)
        self._knfs_helper.get_export_options.assert_any_call(
            self.share, access, 'KNFS')
        cmd = ['exportfs', '-o', export_opts, ':'.join([access['access_to'],
                                                       local_path])]
        self._knfs_helper._publish_access.assert_called_once_with(*cmd)

    def test_knfs_allow_access_access_exists(self):
        out = ['/fs0 <world>', 0]
        self._knfs_helper._execute = mock.Mock(return_value=out)
        self.mock_object(re, 'search', mock.Mock(return_value="fake"))
        self._knfs_helper.get_export_options = mock.Mock()
        access = self.access
        local_path = self.fakesharepath
        self.assertRaises(exception.ShareAccessExists,
                          self._knfs_helper.allow_access,
                          local_path, self.share, access)
        self._knfs_helper._execute.assert_any_call('exportfs',
                                                   run_as_root=True)
        self.assertTrue(re.search.called)
        self.assertFalse(self._knfs_helper.get_export_options.called)

    def test_knfs_allow_access_publish_exception(self):
        self._knfs_helper.get_export_options = mock.Mock()
        self._knfs_helper._publish_access = mock.Mock(
            side_effect=exception.ProcessExecutionError('boom'))

        self.assertRaises(exception.GPFSException,
                          self._knfs_helper.allow_access,
                          self.fakesharepath,
                          self.share,
                          self.access,
                          error_on_exists=False)

        self.assertTrue(self._knfs_helper.get_export_options.called)
        self.assertTrue(self._knfs_helper._publish_access.called)

    def test_knfs_allow_access_invalid_access(self):
        access = fake_share.fake_access(access_type='test')
        self.assertRaises(exception.InvalidShareAccess,
                          self._knfs_helper.allow_access,
                          self.fakesharepath, self.share,
                          access)

    def test_knfs_allow_access_exception(self):
        self._knfs_helper._execute = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )
        access = self.access
        local_path = self.fakesharepath
        self.assertRaises(exception.GPFSException,
                          self._knfs_helper.allow_access,
                          local_path, self.share,
                          access)
        self._knfs_helper._execute.assert_called_once_with('exportfs',
                                                           run_as_root=True)

    def test_knfs__verify_denied_access_pass(self):
        local_path = self.fakesharepath
        ip = self.access['access_to']
        fake_exportfs = ('/shares/share-1\n\t\t1.1.1.1\n'
                         '/shares/share-2\n\t\t2.2.2.2\n')
        self._knfs_helper._publish_access = mock.Mock(
            return_value=[(fake_exportfs, '')])

        self._knfs_helper._verify_denied_access(local_path, self.share, ip)

        self._knfs_helper._publish_access.assert_called_once_with('exportfs')

    def test_knfs__verify_denied_access_fail(self):
        local_path = self.fakesharepath
        ip = self.access['access_to']
        data = {'path': local_path, 'ip': ip}
        fake_exportfs = ('/shares/share-1\n\t\t1.1.1.1\n'
                         '%(path)s\n\t\t%(ip)s\n'
                         '/shares/share-2\n\t\t2.2.2.2\n') % data
        self._knfs_helper._publish_access = mock.Mock(
            return_value=[(fake_exportfs, '')])

        self.assertRaises(exception.GPFSException,
                          self._knfs_helper._verify_denied_access,
                          local_path,
                          self.share,
                          ip)

        self._knfs_helper._publish_access.assert_called_once_with('exportfs')

    def test_knfs__verify_denied_access_exception(self):
        self._knfs_helper._publish_access = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )

        ip = self.access['access_to']
        local_path = self.fakesharepath

        self.assertRaises(exception.GPFSException,
                          self._knfs_helper._verify_denied_access,
                          local_path,
                          self.share,
                          ip)

        self._knfs_helper._publish_access.assert_called_once_with('exportfs')

    @ddt.data((None, False),
              ('', False),
              (' ', False),
              ('Some error to log', True))
    @ddt.unpack
    def test_knfs__verify_denied_access_stderr(self, stderr, is_logged):
        """Stderr debug logging should only happen when not empty."""
        outputs = [('', stderr)]
        self._knfs_helper._publish_access = mock.Mock(return_value=outputs)
        gpfs.LOG.debug = mock.Mock()

        self._knfs_helper._verify_denied_access(
            self.fakesharepath, self.share, self.remote_ip)

        self._knfs_helper._publish_access.assert_called_once_with('exportfs')
        self.assertEqual(is_logged, gpfs.LOG.debug.called)

    def test_knfs_deny_access(self):
        self._knfs_helper._publish_access = mock.Mock(return_value=[('', '')])

        access = self.access
        local_path = self.fakesharepath
        self._knfs_helper.deny_access(local_path, self.share, access)

        deny = ['exportfs', '-u', ':'.join([access['access_to'], local_path])]
        self._knfs_helper._publish_access.assert_has_calls([
            mock.call(*deny, check_exit_code=[0, 1]),
            mock.call('exportfs')])

    def test_knfs_deny_access_exception(self):
        self._knfs_helper._publish_access = mock.Mock(
            side_effect=exception.ProcessExecutionError
        )

        access = self.access
        local_path = self.fakesharepath
        cmd = ['exportfs', '-u', ':'.join([access['access_to'], local_path])]
        self.assertRaises(exception.GPFSException,
                          self._knfs_helper.deny_access, local_path,
                          self.share, access)

        self._knfs_helper._publish_access.assert_called_once_with(
            *cmd, check_exit_code=[0, 1])

    def test_knfs__publish_access(self):
        self.mock_object(utils, 'execute')

        fake_command = 'fakecmd'
        cmd = [fake_command]
        self._knfs_helper._publish_access(*cmd)

        utils.execute.assert_any_call(*cmd, run_as_root=True,
                                      check_exit_code=True)
        remote_login = self.sshlogin + '@' + self.remote_ip
        remote_login2 = self.sshlogin + '@' + self.remote_ip2
        utils.execute.assert_has_calls([
            mock.call('ssh', remote_login, fake_command,
                      check_exit_code=True, run_as_root=False),
            mock.call(fake_command, check_exit_code=True, run_as_root=True),
            mock.call('ssh', remote_login2, fake_command,
                      check_exit_code=True, run_as_root=False)])
        self.assertTrue(socket.gethostbyname_ex.called)
        self.assertTrue(socket.gethostname.called)

    def test_knfs__publish_access_exception(self):
        self.mock_object(
            utils, 'execute',
            mock.Mock(side_effect=(0, exception.ProcessExecutionError)))

        fake_command = 'fakecmd'
        cmd = [fake_command]
        self.assertRaises(exception.ProcessExecutionError,
                          self._knfs_helper._publish_access, *cmd)

        self.assertTrue(socket.gethostbyname_ex.called)
        self.assertTrue(socket.gethostname.called)
        remote_login = self.sshlogin + '@' + self.remote_ip
        utils.execute.assert_has_calls([
            mock.call('ssh', remote_login, fake_command,
                      check_exit_code=True, run_as_root=False),
            mock.call(fake_command, check_exit_code=True, run_as_root=True)])

    @ddt.data('rw', 'ro')
    def test_ces_get_export_options(self, access_level):
        mock_out = {"ces:export_options": "squash=no_root_squash"}
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value=mock_out))
        access = fake_share.fake_access(access_level=access_level)
        out = self._ces_helper.get_export_options(self.share, access, 'CES')
        self.assertEqual("squash=no_root_squash,access_type=%s" % access_level,
                         out)

    def test_ces_get_export_options_default(self):
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value={}))
        access = self.access
        out = self._ces_helper.get_export_options(self.share, access,
                                                  'CES')
        self.assertEqual("access_type=rw", out)

    def test_ces_get_export_options_invalid_option_ro(self):
        mock_out = {"ces:export_options": "access_type=ro"}
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value=mock_out))
        access = self.access
        share = fake_share.fake_share(share_type="fake_share_type")
        self.assertRaises(exception.InvalidInput,
                          self._ces_helper.get_export_options,
                          share, access, 'CES')

    def test_ces_get_export_options_invalid_option_rw(self):
        mock_out = {"ces:export_options": "access_type=rw"}
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value=mock_out))
        access = self.access
        share = fake_share.fake_share(share_type="fake_share_type")
        self.assertRaises(exception.InvalidInput,
                          self._ces_helper.get_export_options,
                          share, access, 'CES')

    def test__get_nfs_client_exports_exception(self):
        self._ces_helper._execute = mock.Mock(return_value=('junk', ''))

        local_path = self.fakesharepath
        self.assertRaises(exception.GPFSException,
                          self._ces_helper._get_nfs_client_exports,
                          local_path)

        self._ces_helper._execute.assert_called_once_with(
            self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n', local_path, '-Y')

    @ddt.data('44.3.2.11', '1:2:3:4:5:6:7:8')
    def test__fix_export_data(self, ip):
        data = None
        for line in self.fake_ces_exports.splitlines():
            if "HEADER" in line:
                headers = line.split(':')
            if ip in line:
                data = line.split(':')
                break
        self.assertIsNotNone(
            data, "Test data did not contain a line with the test IP.")

        result_data = self._ces_helper._fix_export_data(data, headers)

        self.assertEqual(ip, result_data[headers.index('Clients')])

    @ddt.data((None, True),
              ('44.3.2.11', True),
              ('44.3.2.1', False),
              ('4.3.2.1', False),
              ('4.3.2.11', False),
              ('1.2.3.4', False),
              ('', False),
              ('*', False),
              ('.', False),
              ('1:2:3:4:5:6:7:8', True))
    @ddt.unpack
    def test_ces__has_client_access(self, ip, has_access):
        mock_out = self.fake_ces_exports
        self._ces_helper._execute = mock.Mock(
            return_value=(mock_out, ''))

        local_path = self.fakesharepath
        self.assertEqual(has_access,
                         self._ces_helper._has_client_access(local_path, ip))

        self._ces_helper._execute.assert_called_once_with(
            self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n', local_path, '-Y')

    def test_ces_remove_export_no_exports(self):
        mock_out = self.fake_ces_exports_not_found
        self._ces_helper._execute = mock.Mock(
            return_value=(mock_out, ''))

        local_path = self.fakesharepath
        self._ces_helper.remove_export(local_path, self.share)

        self._ces_helper._execute.assert_called_once_with(
            self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n', local_path, '-Y')

    def test_ces_remove_export_existing_exports(self):
        mock_out = self.fake_ces_exports
        self._ces_helper._execute = mock.Mock(
            return_value=(mock_out, ''))

        local_path = self.fakesharepath
        self._ces_helper.remove_export(local_path, self.share)

        self._ces_helper._execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n',
                      local_path, '-Y'),
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'remove',
                      local_path),
        ])

    def test_ces_remove_export_exception(self):
        local_path = self.fakesharepath
        self._ces_helper._execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.GPFSException,
                          self._ces_helper.remove_export,
                          local_path, self.share)

    def test_ces_allow_access(self):
        mock_out = self.fake_ces_exports_not_found
        self._ces_helper._execute = mock.Mock(
            return_value=(mock_out, ''))

        export_opts = "access_type=rw"
        self._ces_helper.get_export_options = mock.Mock(
            return_value=export_opts)

        access = self.access
        local_path = self.fakesharepath

        self._ces_helper.allow_access(local_path, self.share, access)

        self._ces_helper._execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n',
                      local_path, '-Y'),
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'add', local_path,
                      '-c', access['access_to'] + '(' + export_opts + ')')])

    def test_ces_allow_access_existing_exports(self):
        mock_out = self.fake_ces_exports
        self._ces_helper._execute = mock.Mock(
            return_value=(mock_out, ''))

        export_opts = "access_type=rw"
        self._ces_helper.get_export_options = mock.Mock(
            return_value=export_opts)

        access = self.access
        local_path = self.fakesharepath

        self._ces_helper.allow_access(self.fakesharepath, self.share,
                                      self.access)

        self._ces_helper._execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n',
                      local_path, '-Y'),
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'change', local_path,
                      '--nfsadd', access['access_to'] + '(' +
                      export_opts + ')')])

    def test_ces_allow_access_invalid_access_type(self):
        access = fake_share.fake_access(access_type='test')
        self.assertRaises(exception.InvalidShareAccess,
                          self._ces_helper.allow_access,
                          self.fakesharepath, self.share,
                          access)

    def test_ces_allow_access_exception(self):
        access = self.access
        local_path = self.fakesharepath
        self._ces_helper._execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.GPFSException,
                          self._ces_helper.allow_access, local_path,
                          self.share, access)

    def test_ces_deny_access(self):
        mock_out = self.fake_ces_exports
        self._ces_helper._execute = mock.Mock(
            return_value=(mock_out, ''))

        access = self.access
        local_path = self.fakesharepath

        self._ces_helper.deny_access(local_path, self.share, access)

        self._ces_helper._execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n',
                      local_path, '-Y'),
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'change', local_path,
                      '--nfsremove', access['access_to'])])

    def test_ces_deny_access_exception(self):
        access = self.access
        local_path = self.fakesharepath
        self._ces_helper._execute = mock.Mock(
            side_effect=exception.ProcessExecutionError)
        self.assertRaises(exception.GPFSException,
                          self._ces_helper.deny_access, local_path,
                          self.share, access)

    def test_ces_resync_access_add(self):
        mock_out = self.fake_ces_exports_not_found
        self._ces_helper._execute = mock.Mock(return_value=(mock_out, ''))
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value={}))

        access_rules = [self.access]
        local_path = self.fakesharepath
        self._ces_helper.resync_access(local_path, self.share, access_rules)

        self._ces_helper._execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n',
                      local_path, '-Y'),
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'add', local_path,
                      '-c', self.access['access_to'] + '(' + "access_type=rw" +
                      ')')
        ])
        share_types.get_extra_specs_from_share.assert_called_once_with(
            self.share)

    def test_ces_resync_access_change(self):

        class SortedMatch(object):
            def __init__(self, f, expected):
                self.assertEqual = f
                self.expected = expected

            def __eq__(self, actual):
                expected_list = self.expected.split(',')
                actual_list = actual.split(',')
                self.assertEqual(sorted(expected_list), sorted(actual_list))
                return True

        mock_out = self.fake_ces_exports
        self._ces_helper._execute = mock.Mock(
            return_value=(mock_out, ''))
        self.mock_object(share_types, 'get_extra_specs_from_share',
                         mock.Mock(return_value={}))

        access_rules = [fake_share.fake_access(access_to='1.1.1.1'),
                        fake_share.fake_access(
                            access_to='10.0.0.1', access_level='ro')]
        local_path = self.fakesharepath
        self._ces_helper.resync_access(local_path, self.share, access_rules)

        share_types.get_extra_specs_from_share.assert_called_once_with(
            self.share)
        to_remove = '1:2:3:4:5:6:7:8,44.3.2.11'
        to_add = access_rules[0]['access_to'] + '(' + "access_type=rw" + ')'
        to_change = access_rules[1]['access_to'] + '(' + "access_type=ro" + ')'
        self._ces_helper._execute.assert_has_calls([
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n',
                      local_path, '-Y'),
            mock.call(self.GPFS_PATH + 'mmnfs', 'export', 'change', local_path,
                      '--nfsremove', SortedMatch(self.assertEqual, to_remove),
                      '--nfsadd', to_add,
                      '--nfschange', to_change)
        ])

    def test_ces_resync_nothing(self):
        """Test that hits the add-no-rules case."""
        mock_out = self.fake_ces_exports_not_found
        self._ces_helper._execute = mock.Mock(return_value=(mock_out, ''))

        local_path = self.fakesharepath
        self._ces_helper.resync_access(local_path, None, [])

        self._ces_helper._execute.assert_called_once_with(
            self.GPFS_PATH + 'mmnfs', 'export', 'list', '-n', local_path, '-Y')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Abstract base class for NetApp NAS protocol helper classes.
"""

import abc

import six

from manila.common import constants
from manila import utils


def access_rules_synchronized(f):
    """Decorator for synchronizing share access rule modification methods."""

    def wrapped_func(self, *args, **kwargs):

        # The first argument is always a share, which has an ID
        key = "share-access-%s" % args[0]['id']

        @utils.synchronized(key)
        def source_func(self, *args, **kwargs):
            return f(self, *args, **kwargs)

        return source_func(self, *args, **kwargs)

    return wrapped_func


@six.add_metaclass(abc.ABCMeta)
class NetAppBaseHelper(object):
    """Interface for protocol-specific NAS drivers."""

    def __init__(self):
        self._client = None

    def set_client(self, client):
        self._client = client

    def _is_readonly(self, access_level):
        """Returns whether an access rule specifies read-only access."""
        return access_level == constants.ACCESS_LEVEL_RO

    @staticmethod
    def _get_share_export_location(share):
        """Returns the export location of the share.

        The share may contain only the list of export location, depending on
        the entity provided by the manager.
        """
        export_location = share.get('export_location')
        if not export_location:
            export_location_list = share.get('export_locations')
            if (isinstance(export_location_list, list) and
                    len(export_location_list) > 0):
                export_location = export_location_list[0]['path']

        return export_location

    @abc.abstractmethod
    def create_share(self, share, share_name,
                     clear_current_export_policy=True,
                     ensure_share_already_exists=False, replica=False,
                     is_flexgroup=False):
        """Creates NAS share."""

    @abc.abstractmethod
    def delete_share(self, share, share_name):
        """Deletes NAS share."""

    @abc.abstractmethod
    def update_access(self, share, share_name, rules):
        """Replaces the list of access rules known to the backend storage."""

    @abc.abstractmethod
    def get_target(self, share):
        """Returns host where the share located."""

    @abc.abstractmethod
    def get_share_name_for_share(self, share):
        """Returns the flexvol name that hosts a share."""

    @abc.abstractmethod
    def cleanup_demoted_replica(self, share, share_name):
        """Do some cleanup regarding the former active replica"""



# Copyright 2013 OpenStack Foundation
# Copyright 2014 Mirantis Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from neutronclient.common import exceptions as neutron_client_exc
from neutronclient.v2_0 import client as clientv20
from oslo_config import cfg
from oslo_log import log

from manila.common import client_auth
from manila import context
from manila import exception
from manila.network.neutron import constants as neutron_constants

NEUTRON_GROUP = 'neutron'


neutron_opts = [
    cfg.StrOpt(
        'url',
        default='http://127.0.0.1:9696',
        help='URL for connecting to neutron.'),
    cfg.IntOpt(
        'url_timeout',
        default=30,
        help='Timeout value for connecting to neutron in seconds.'),
    cfg.StrOpt(
        'auth_strategy',
        default='keystone',
        help='Auth strategy for connecting to neutron in admin context.'),
    cfg.StrOpt(
        'endpoint_type',
        default='publicURL',
        help='Endpoint type to be used with neutron client calls.'),
    cfg.StrOpt(
        'region_name',
        help='Region name for connecting to neutron in admin context.'),
]


CONF = cfg.CONF
LOG = log.getLogger(__name__)


def list_opts():
    return client_auth.AuthClientLoader.list_opts(NEUTRON_GROUP)


class API(object):
    """API for interacting with the neutron 2.x API.

    :param configuration: instance of config or config group.
    """

    def __init__(self, config_group_name=None):
        self.config_group_name = config_group_name or 'DEFAULT'

        ks_loading.register_session_conf_options(
            CONF, NEUTRON_GROUP)
        ks_loading.register_auth_conf_options(CONF, NEUTRON_GROUP)
        CONF.register_opts(neutron_opts, NEUTRON_GROUP)

        self.configuration = getattr(CONF, self.config_group_name, CONF)
        self.last_neutron_extension_sync = None
        self.extensions = {}
        self.auth_obj = None

    @property
    def client(self):
        return self.get_client(context.get_admin_context())

    def get_client(self, context):
        if not self.auth_obj:
            self.auth_obj = client_auth.AuthClientLoader(
                client_class=clientv20.Client, cfg_group=NEUTRON_GROUP)

        return self.auth_obj.get_client(
            self,
            context,
            endpoint_type=CONF[NEUTRON_GROUP].endpoint_type,
            region_name=CONF[NEUTRON_GROUP].region_name,
        )

    @property
    def admin_project_id(self):
        if self.client.httpclient.auth_token is None:
            try:
                self.client.httpclient.authenticate()
            except neutron_client_exc.NeutronClientException as e:
                raise exception.NetworkException(code=e.status_code,
                                                 message=e.message)
        return self.client.httpclient.get_project_id()

    def get_all_admin_project_networks(self):
        search_opts = {'tenant_id': self.admin_project_id, 'shared': False}
        nets = self.client.list_networks(**search_opts).get('networks', [])
        return nets

    def create_port(self, tenant_id, network_id, host_id=None, subnet_id=None,
                    fixed_ip=None, device_owner=None, device_id=None,
                    mac_address=None, port_security_enabled=True,
                    security_group_ids=None, dhcp_opts=None, **kwargs):
        try:
            port_req_body = {'port': {}}
            port_req_body['port']['network_id'] = network_id
            port_req_body['port']['admin_state_up'] = True
            port_req_body['port']['tenant_id'] = tenant_id
            if not port_security_enabled:
                port_req_body['port']['port_security_enabled'] = (
                    port_security_enabled)
            elif security_group_ids:
                port_req_body['port']['security_groups'] = security_group_ids
            if mac_address:
                port_req_body['port']['mac_address'] = mac_address
            if host_id:
                if not self._has_port_binding_extension():
                    msg = ("host_id (%(host_id)s) specified but neutron "
                           "doesn't support port binding. Please activate the "
                           "extension accordingly." % {"host_id": host_id})
                    raise exception.NetworkException(message=msg)
                port_req_body['port']['binding:host_id'] = host_id
            if dhcp_opts is not None:
                port_req_body['port']['extra_dhcp_opts'] = dhcp_opts
            if subnet_id:
                fixed_ip_dict = {'subnet_id': subnet_id}
                if fixed_ip:
                    fixed_ip_dict.update({'ip_address': fixed_ip})
                port_req_body['port']['fixed_ips'] = [fixed_ip_dict]
            if device_owner:
                port_req_body['port']['device_owner'] = device_owner
            if device_id:
                port_req_body['port']['device_id'] = device_id
            if kwargs:
                port_req_body['port'].update(kwargs)
            port = self.client.create_port(port_req_body).get('port', {})
            return port
        except neutron_client_exc.NeutronClientException as e:
            LOG.exception('Neutron error creating port on network %s',
                          network_id)
            if e.status_code == 409:
                raise exception.PortLimitExceeded()
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def delete_port(self, port_id):
        try:
            self.client.delete_port(port_id)
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def delete_subnet(self, subnet_id):
        try:
            self.client.delete_subnet(subnet_id)
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def list_ports(self, **search_opts):
        """List ports for the client based on search options."""
        return self.client.list_ports(**search_opts).get('ports')

    def show_port(self, port_id):
        """Return the port for the client given the port id."""
        try:
            return self.client.show_port(port_id).get('port')
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def get_all_networks(self):
        """Get all networks for client."""
        return self.client.list_networks().get('networks')

    def get_network(self, network_uuid):
        """Get specific network for client."""
        try:
            network = self.client.show_network(network_uuid).get('network', {})
            return network
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def get_subnet(self, subnet_uuid):
        """Get specific subnet for client."""
        try:
            return self.client.show_subnet(subnet_uuid).get('subnet', {})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def list_extensions(self):
        extensions_list = self.client.list_extensions().get('extensions')
        return {ext['name']: ext for ext in extensions_list}

    def _has_port_binding_extension(self):
        if not self.extensions:
            self.extensions = self.list_extensions()
        return neutron_constants.PORTBINDING_EXT in self.extensions

    def router_create(self, tenant_id, name):
        router_req_body = {'router': {}}
        router_req_body['router']['tenant_id'] = tenant_id
        router_req_body['router']['name'] = name
        try:
            return self.client.create_router(router_req_body).get('router', {})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def network_create(self, tenant_id, name):
        network_req_body = {'network': {}}
        network_req_body['network']['tenant_id'] = tenant_id
        network_req_body['network']['name'] = name
        try:
            return self.client.create_network(
                network_req_body).get('network', {})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def subnet_create(self, tenant_id, net_id, name, cidr):
        subnet_req_body = {'subnet': {}}
        subnet_req_body['subnet']['tenant_id'] = tenant_id
        subnet_req_body['subnet']['name'] = name
        subnet_req_body['subnet']['network_id'] = net_id
        subnet_req_body['subnet']['cidr'] = cidr
        subnet_req_body['subnet']['ip_version'] = 4
        try:
            return self.client.create_subnet(
                subnet_req_body).get('subnet', {})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def router_add_interface(self, router_id, subnet_id, port_id=None):
        body = {}
        if subnet_id:
            body['subnet_id'] = subnet_id
        if port_id:
            body['port_id'] = port_id
        try:
            self.client.add_interface_router(router_id, body)
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def router_remove_interface(self, router_id, subnet_id, port_id=None):
        body = {}
        if subnet_id:
            body['subnet_id'] = subnet_id
        if port_id:
            body['port_id'] = port_id
        try:
            self.client.remove_interface_router(router_id, body)
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def router_list(self):
        try:
            return self.client.list_routers().get('routers', {})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def update_port_fixed_ips(self, port_id, fixed_ips):
        try:
            port_req_body = {'port': fixed_ips}
            port = self.client.update_port(
                port_id, port_req_body).get('port', {})
            return port
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def show_router(self, router_id):
        try:
            return self.client.show_router(router_id).get('router', {})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def router_update_routes(self, router_id, routes):
        try:
            router_req_body = {'router': routes}
            port = self.client.update_router(
                router_id, router_req_body).get('router', {})
            return port
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def update_subnet(self, subnet_uuid, name):
        """Update specific subnet for client."""
        subnet_req_body = {'subnet': {'name': name}}
        try:
            return self.client.update_subnet(
                subnet_uuid, subnet_req_body).get('subnet', {})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(code=e.status_code,
                                             message=e.message)

    def security_group_list(self, search_opts=None):
        try:
            return self.client.list_security_groups(**search_opts)
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(
                code=e.status_code, message=e.message)

    def security_group_create(self, name, description=""):
        try:
            return self.client.create_security_group(
                {'security_group': {"name": name, "description": description}})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(
                code=e.status_code, message=e.message)

    def security_group_rule_create(self, parent_group_id,
                                   ip_protocol=None, from_port=None,
                                   to_port=None, cidr=None, group_id=None,
                                   direction="ingress"):
        request = {"security_group_id": parent_group_id,
                   "protocol": ip_protocol, "remote_ip_prefix": cidr,
                   "remote_group_id": group_id, "direction": direction}
        if ip_protocol != "icmp":
            request["port_range_min"] = from_port
            request["port_range_max"] = to_port

        try:
            return self.client.create_security_group_rule(
                {"security_group_rule": request})
        except neutron_client_exc.NeutronClientException as e:
            raise exception.NetworkException(
                code=e.status_code, message=e.message)



# Copyright (c) 2016 by Tegile Systems, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Share driver for Tegile storage.
"""

import json
import requests
import six

from oslo_config import cfg
from oslo_log import log

from manila import exception
from manila.i18n import _
from manila.share import driver
from manila.share import utils as share_utils
from manila import utils

tegile_opts = [
    cfg.HostAddressOpt('tegile_nas_server',
                       help='Tegile NAS server hostname or IP address.'),
    cfg.StrOpt('tegile_nas_login',
               help='User name for the Tegile NAS server.'),
    cfg.StrOpt('tegile_nas_password',
               help='Password for the Tegile NAS server.'),
    cfg.StrOpt('tegile_default_project',
               help='Create shares in this project')]


CONF = cfg.CONF
CONF.register_opts(tegile_opts)

LOG = log.getLogger(__name__)
DEFAULT_API_SERVICE = 'openstack'
TEGILE_API_PATH = 'zebi/api'
TEGILE_LOCAL_CONTAINER_NAME = 'Local'
TEGILE_SNAPSHOT_PREFIX = 'Manual-S-'
VENDOR = 'Tegile Systems Inc.'
DEFAULT_BACKEND_NAME = 'Tegile'
VERSION = '1.0.0'
DEBUG_LOGGING = False  # For debugging purposes


def debugger(func):
    """Returns a wrapper that wraps func.

    The wrapper will log the entry and exit points of the function.
    """

    def wrapper(*args, **kwds):
        if DEBUG_LOGGING:
            LOG.debug('Entering %(classname)s.%(funcname)s',
                      {
                          'classname': args[0].__class__.__name__,
                          'funcname': func.__name__,
                      })
            LOG.debug('Arguments: %(args)s, %(kwds)s',
                      {
                          'args': args[1:],
                          'kwds': kwds,
                      })
        f_result = func(*args, **kwds)
        if DEBUG_LOGGING:
            LOG.debug('Exiting %(classname)s.%(funcname)s',
                      {
                          'classname': args[0].__class__.__name__,
                          'funcname': func.__name__,
                      })
            LOG.debug('Results: %(result)s',
                      {'result': f_result})
        return f_result

    return wrapper


class TegileAPIExecutor(object):
    def __init__(self, classname, hostname, username, password):
        self._classname = classname
        self._hostname = hostname
        self._username = username
        self._password = password

    def __call__(self, *args, **kwargs):
        return self._send_api_request(*args, **kwargs)

    @debugger
    @utils.retry(retry_param=(requests.ConnectionError, requests.Timeout),
                 interval=30,
                 retries=3,
                 backoff_rate=1)
    def _send_api_request(self, method, params=None,
                          request_type='post',
                          api_service=DEFAULT_API_SERVICE,
                          fine_logging=DEBUG_LOGGING):
        if params is not None:
            params = json.dumps(params)

        url = 'https://%s/%s/%s/%s' % (self._hostname,
                                       TEGILE_API_PATH,
                                       api_service,
                                       method)
        if fine_logging:
            LOG.debug('TegileAPIExecutor(%(classname)s) method: %(method)s, '
                      'url: %(url)s', {
                          'classname': self._classname,
                          'method': method,
                          'url': url,
                      })
        if request_type == 'post':
            if fine_logging:
                LOG.debug('TegileAPIExecutor(%(classname)s) '
                          'method: %(method)s, payload: %(payload)s',
                          {
                              'classname': self._classname,
                              'method': method,
                              'payload': params,
                          })
            req = requests.post(url,
                                data=params,
                                auth=(self._username, self._password),
                                verify=False)
        else:
            req = requests.get(url,
                               auth=(self._username, self._password),
                               verify=False)

        if fine_logging:
            LOG.debug('TegileAPIExecutor(%(classname)s) method: %(method)s, '
                      'return code: %(retcode)s',
                      {
                          'classname': self._classname,
                          'method': method,
                          'retcode': req,
                      })
        try:
            response = req.json()
            if fine_logging:
                LOG.debug('TegileAPIExecutor(%(classname)s) '
                          'method: %(method)s, response: %(response)s',
                          {
                              'classname': self._classname,
                              'method': method,
                              'response': response,
                          })
        except ValueError:
            # Some APIs don't return output and that's fine
            response = ''
        req.close()

        if req.status_code != 200:
            raise exception.TegileAPIException(response=req.text)

        return response


class TegileShareDriver(driver.ShareDriver):
    """Tegile NAS driver. Allows for NFS and CIFS NAS storage usage."""
    def __init__(self, *args, **kwargs):
        super(TegileShareDriver, self).__init__(False, *args, **kwargs)

        self.configuration.append_config_values(tegile_opts)
        self._default_project = (self.configuration.safe_get(
            "tegile_default_project") or 'openstack')
        self._backend_name = (self.configuration.safe_get('share_backend_name')
                              or CONF.share_backend_name
                              or DEFAULT_BACKEND_NAME)
        self._hostname = self.configuration.safe_get('tegile_nas_server')
        username = self.configuration.safe_get('tegile_nas_login')
        password = self.configuration.safe_get('tegile_nas_password')
        self._api = TegileAPIExecutor(self.__class__.__name__,
                                      self._hostname,
                                      username,
                                      password)

    @debugger
    def create_share(self, context, share, share_server=None):
        """Is called to create share."""
        share_name = share['name']
        share_proto = share['share_proto']

        pool_name = share_utils.extract_host(share['host'], level='pool')

        params = (pool_name, self._default_project, share_name, share_proto)

        # Share name coming from the backend is the most reliable. Sometimes
        # a few options in Tegile array could cause sharename to be different
        # from the one passed to it. Eg. 'projectname-sharename' instead
        # of 'sharename' if inherited share properties are selected.
        ip, real_share_name = self._api('createShare', params).split()

        LOG.info("Created share %(sharename)s, share id %(shid)s.",
                 {'sharename': share_name, 'shid': share['id']})

        return self._get_location_path(real_share_name, share_proto, ip)

    @debugger
    def extend_share(self, share, new_size, share_server=None):
        """Is called to extend share.

        There is no resize for Tegile shares.
        We just adjust the quotas. The API is still called 'resizeShare'.
        """

        self._adjust_size(share, new_size, share_server)

    @debugger
    def shrink_share(self, shrink_share, shrink_size, share_server=None):
        """Uses resize_share to shrink a share.

        There is no shrink for Tegile shares.
        We just adjust the quotas. The API is still called 'resizeShare'.
        """
        self._adjust_size(shrink_share, shrink_size, share_server)

    @debugger
    def _adjust_size(self, share, new_size, share_server=None):
        pool, project, share_name = self._get_pool_project_share_name(share)
        params = ('%s/%s/%s/%s' % (pool,
                                   TEGILE_LOCAL_CONTAINER_NAME,
                                   project,
                                   share_name),
                  six.text_type(new_size),
                  'GB')
        self._api('resizeShare', params)

    @debugger
    def delete_share(self, context, share, share_server=None):
        """Is called to remove share."""
        pool, project, share_name = self._get_pool_project_share_name(share)
        params = ('%s/%s/%s/%s' % (pool,
                                   TEGILE_LOCAL_CONTAINER_NAME,
                                   project,
                                   share_name),
                  True,
                  False)

        self._api('deleteShare', params)

    @debugger
    def create_snapshot(self, context, snapshot, share_server=None):
        """Is called to create snapshot."""
        snap_name = snapshot['name']

        pool, project, share_name = self._get_pool_project_share_name(
            snapshot['share'])

        share = {
            'poolName': '%s' % pool,
            'projectName': '%s' % project,
            'name': share_name,
            'availableSize': 0,
            'totalSize': 0,
            'datasetPath': '%s/%s/%s' %
                           (pool,
                            TEGILE_LOCAL_CONTAINER_NAME,
                            project),
            'mountpoint': share_name,
            'local': 'true',
        }

        params = (share, snap_name, False)

        LOG.info('Creating snapshot for share_name=%(shr)s'
                 ' snap_name=%(name)s',
                 {'shr': share_name, 'name': snap_name})

        self._api('createShareSnapshot', params)

    @debugger
    def create_share_from_snapshot(self, context, share, snapshot,
                                   share_server=None, parent_share=None):
        """Create a share from a snapshot - clone a snapshot."""
        pool, project, share_name = self._get_pool_project_share_name(share)

        params = ('%s/%s/%s/%s@%s%s' % (pool,
                                        TEGILE_LOCAL_CONTAINER_NAME,
                                        project,
                                        snapshot['share_name'],
                                        TEGILE_SNAPSHOT_PREFIX,
                                        snapshot['name'],
                                        ),
                  share_name,
                  True,
                  )

        ip, real_share_name = self._api('cloneShareSnapshot',
                                        params).split()

        share_proto = share['share_proto']
        return self._get_location_path(real_share_name, share_proto, ip)

    @debugger
    def delete_snapshot(self, context, snapshot, share_server=None):
        """Is called to remove snapshot."""
        pool, project, share_name = self._get_pool_project_share_name(
            snapshot['share'])
        params = ('%s/%s/%s/%s@%s%s' % (pool,
                                        TEGILE_LOCAL_CONTAINER_NAME,
                                        project,
                                        share_name,
                                        TEGILE_SNAPSHOT_PREFIX,
                                        snapshot['name']),
                  False)

        self._api('deleteShareSnapshot', params)

    @debugger
    def ensure_share(self, context, share, share_server=None):
        """Invoked to sure that share is exported."""

        # Fetching share name from server, because some configuration
        # options can cause sharename different from the OpenStack share name
        pool, project, share_name = self._get_pool_project_share_name(share)
        params = [
            '%s/%s/%s/%s' % (pool,
                             TEGILE_LOCAL_CONTAINER_NAME,
                             project,
                             share_name),
        ]
        ip, real_share_name = self._api('getShareIPAndMountPoint',
                                        params).split()

        share_proto = share['share_proto']
        location = self._get_location_path(real_share_name, share_proto, ip)
        return [location]

    @debugger
    def _allow_access(self, context, share, access, share_server=None):
        """Allow access to the share."""
        share_proto = share['share_proto']
        access_type = access['access_type']
        access_level = access['access_level']
        access_to = access['access_to']

        self._check_share_access(share_proto, access_type)

        pool, project, share_name = self._get_pool_project_share_name(share)
        params = ('%s/%s/%s/%s' % (pool,
                                   TEGILE_LOCAL_CONTAINER_NAME,
                                   project,
                                   share_name),
                  share_proto,
                  access_type,
                  access_to,
                  access_level)

        self._api('shareAllowAccess', params)

    @debugger
    def _deny_access(self, context, share, access, share_server=None):
        """Deny access to the share."""
        share_proto = share['share_proto']
        access_type = access['access_type']
        access_level = access['access_level']
        access_to = access['access_to']

        self._check_share_access(share_proto, access_type)

        pool, project, share_name = self._get_pool_project_share_name(share)
        params = ('%s/%s/%s/%s' % (pool,
                                   TEGILE_LOCAL_CONTAINER_NAME,
                                   project,
                                   share_name),
                  share_proto,
                  access_type,
                  access_to,
                  access_level)

        self._api('shareDenyAccess', params)

    def _check_share_access(self, share_proto, access_type):
        if share_proto == 'CIFS' and access_type != 'user':
            reason = ('Only USER access type is allowed for '
                      'CIFS shares.')
            LOG.warning(reason)
            raise exception.InvalidShareAccess(reason=reason)
        elif share_proto == 'NFS' and access_type not in ('ip', 'user'):
            reason = ('Only IP or USER access types are allowed for '
                      'NFS shares.')
            LOG.warning(reason)
            raise exception.InvalidShareAccess(reason=reason)
        elif share_proto not in ('NFS', 'CIFS'):
            reason = ('Unsupported protocol \"%s\" specified for '
                      'access rule.') % share_proto
            raise exception.InvalidShareAccess(reason=reason)

    @debugger
    def update_access(self, context, share, access_rules, add_rules,
                      delete_rules, share_server=None):
        if not (add_rules or delete_rules):
            # Recovery mode
            pool, project, share_name = (
                self._get_pool_project_share_name(share))
            share_proto = share['share_proto']
            params = ('%s/%s/%s/%s' % (pool,
                                       TEGILE_LOCAL_CONTAINER_NAME,
                                       project,
                                       share_name),
                      share_proto)

            # Clears all current ACLs
            # Remove ip and user ACLs if share_proto is NFS
            # Remove user ACLs if share_proto is CIFS
            self._api('clearAccessRules', params)

            # Looping through all rules.
            # Will have one API call per rule.
            for access in access_rules:
                self._allow_access(context, share, access, share_server)
        else:
            # Adding/Deleting specific rules
            for access in delete_rules:
                self._deny_access(context, share, access, share_server)
            for access in add_rules:
                self._allow_access(context, share, access, share_server)

    @debugger
    def _update_share_stats(self, **kwargs):
        """Retrieve stats info."""

        try:
            data = self._api(method='getArrayStats',
                             request_type='get',
                             fine_logging=False)
            # fixing values coming back here as String to float
            for pool in data.get('pools', []):
                pool['total_capacity_gb'] = float(
                    pool.get('total_capacity_gb', 0))
                pool['free_capacity_gb'] = float(
                    pool.get('free_capacity_gb', 0))
                pool['allocated_capacity_gb'] = float(
                    pool.get('allocated_capacity_gb', 0))

                pool['qos'] = pool.pop('QoS_support', False)
                pool['reserved_percentage'] = (
                    self.configuration.reserved_share_percentage)
                pool['reserved_snapshot_percentage'] = (
                    self.configuration.reserved_share_from_snapshot_percentage
                    or self.configuration.reserved_share_percentage)
                pool['dedupe'] = True
                pool['compression'] = True
                pool['thin_provisioning'] = True
                pool['max_over_subscription_ratio'] = (
                    self.configuration.max_over_subscription_ratio)

            data['share_backend_name'] = self._backend_name
            data['vendor_name'] = VENDOR
            data['driver_version'] = VERSION
            data['storage_protocol'] = 'NFS_CIFS'
            data['snapshot_support'] = True
            data['create_share_from_snapshot_support'] = True
            data['qos'] = False

            super(TegileShareDriver, self)._update_share_stats(data)
        except Exception:
            msg = _('Unexpected error while trying to get the '
                    'usage stats from array.')
            LOG.exception(msg)
            raise

    @debugger
    def get_pool(self, share):
        """Returns pool name where share resides.

        :param share: The share hosted by the driver.
        :return: Name of the pool where given share is hosted.
        """
        pool = share_utils.extract_host(share['host'], level='pool')
        return pool

    @debugger
    def get_network_allocations_number(self):
        """Get number of network interfaces to be created."""
        return 0

    @debugger
    def _get_location_path(self, share_name, share_proto, ip=None):
        if ip is None:
            ip = self._hostname
        if share_proto == 'NFS':
            location = '%s:%s' % (ip, share_name)
        elif share_proto == 'CIFS':
            location = r'\\%s\%s' % (ip, share_name)
        else:
            message = _('Invalid NAS protocol supplied: %s.') % share_proto
            raise exception.InvalidInput(message)

        export_location = {
            'path': location,
            'is_admin_only': False,
            'metadata': {
                'preferred': True,
            },
        }
        return export_location

    @debugger
    def _get_pool_project_share_name(self, share):
        pool = share_utils.extract_host(share['host'], level='pool')
        project = self._default_project

        share_name = share['name']

        return pool, project, share_name



# Copyright 2015 Mirantis inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock

import ddt
from oslo_utils import uuidutils

from manila.share import drivers_private_data as pd
from manila import test


@ddt.ddt
class DriverPrivateDataTestCase(test.TestCase):
    """Tests DriverPrivateData."""

    def setUp(self):
        super(DriverPrivateDataTestCase, self).setUp()

        self.fake_storage = mock.Mock()
        self.entity_id = uuidutils.generate_uuid()

    def test_default_storage_driver(self):
        private_data = pd.DriverPrivateData(
            storage=None, context="fake", backend_host="fake")

        self.assertIsInstance(private_data._storage, pd.SqlStorageDriver)

    def test_custom_storage_driver(self):
        private_data = pd.DriverPrivateData(storage=self.fake_storage)

        self.assertEqual(self.fake_storage, private_data._storage)

    def test_invalid_parameters(self):
        self.assertRaises(ValueError, pd.DriverPrivateData)

    @ddt.data({'context': 'fake'}, {'backend_host': 'fake'})
    def test_invalid_single_parameter(self, test_args):
        self.assertRaises(ValueError, pd.DriverPrivateData, **test_args)

    @ddt.data("111", ["fake"], None)
    def test_validate_entity_id_invalid(self, entity_id):
        data = pd.DriverPrivateData(storage="fake")

        self.assertRaises(ValueError, data._validate_entity_id, entity_id)

    def test_validate_entity_id_valid(self):
        actual_result = (
            pd.DriverPrivateData._validate_entity_id(self.entity_id)
        )

        self.assertIsNone(actual_result)

    def test_update(self):
        data = pd.DriverPrivateData(storage=self.fake_storage)
        details = {"foo": "bar"}
        self.mock_object(self.fake_storage, 'update',
                         mock.Mock(return_value=True))

        actual_result = data.update(
            self.entity_id,
            details,
            delete_existing=True
        )

        self.assertTrue(actual_result)
        self.fake_storage.update.assert_called_once_with(
            self.entity_id, details, True
        )

    def test_update_invalid(self):
        data = pd.DriverPrivateData(storage=self.fake_storage)
        details = ["invalid"]
        self.mock_object(self.fake_storage, 'update',
                         mock.Mock(return_value=True))

        self.assertRaises(
            ValueError, data.update, self.entity_id, details)

        self.assertFalse(self.fake_storage.update.called)

    def test_get(self):
        data = pd.DriverPrivateData(storage=self.fake_storage)
        key = "fake_key"
        value = "fake_value"
        default_value = "def"
        self.mock_object(self.fake_storage, 'get',
                         mock.Mock(return_value=value))

        actual_result = data.get(self.entity_id, key, default_value)

        self.assertEqual(value, actual_result)
        self.fake_storage.get.assert_called_once_with(
            self.entity_id, key, default_value
        )

    def test_delete(self):
        data = pd.DriverPrivateData(storage=self.fake_storage)
        key = "fake_key"
        self.mock_object(self.fake_storage, 'get',
                         mock.Mock(return_value=True))

        actual_result = data.delete(self.entity_id, key)

        self.assertTrue(actual_result)
        self.fake_storage.delete.assert_called_once_with(
            self.entity_id, key
        )


fake_storage_data = {
    "entity_id": "fake_id",
    "details": {"foo": "bar"},
    "context": "fake_context",
    "backend_host": "fake_host",
    "default": "def",
    "delete_existing": True,
    "key": "fake_key",
}


def create_arg_list(key_names):
    return [fake_storage_data[key] for key in key_names]


def create_arg_dict(key_names):
    return {key: fake_storage_data[key] for key in key_names}


@ddt.ddt
class SqlStorageDriverTestCase(test.TestCase):

    @ddt.data(
        {
            "method_name": 'update',
            "method_kwargs": create_arg_dict(
                ["entity_id", "details", "delete_existing"]),
            "valid_args": create_arg_list(
                ["context", "entity_id", "details",
                 "delete_existing"]
            )
        },
        {
            "method_name": 'get',
            "method_kwargs": create_arg_dict(["entity_id", "key", "default"]),
            "valid_args": create_arg_list(
                ["context", "entity_id", "key", "default"]),
        },
        {
            "method_name": 'delete',
            "method_kwargs": create_arg_dict(["entity_id", "key"]),
            "valid_args": create_arg_list(
                ["context", "entity_id", "key"]),
        })
    @ddt.unpack
    def test_methods(self, method_kwargs, method_name, valid_args):
        method = method_name
        db_method = 'driver_private_data_' + method_name

        with mock.patch('manila.db.api.' + db_method) as db_method:
            storage_driver = pd.SqlStorageDriver(
                context=fake_storage_data['context'],
                backend_host=fake_storage_data['backend_host'])
            method = getattr(storage_driver, method)

            method(**method_kwargs)

            db_method.assert_called_once_with(*valid_args)



# Copyright 2011 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Fakes For Scheduler tests.
"""

from oslo_utils import timeutils

from manila.scheduler.drivers import filter
from manila.scheduler import host_manager
from manila.scheduler.weighers import base_host as base_host_weigher


FAKE_AZ_1 = {'name': 'zone1', 'id': '24433438-c9b5-4cb5-a472-f78462aa5f31'}
FAKE_AZ_2 = {'name': 'zone2', 'id': 'ebef050c-d20d-4c44-b272-1a0adce11cb5'}
FAKE_AZ_3 = {'name': 'zone3', 'id': '18e7e6e2-39d6-466b-a706-2717bd1086e1'}
FAKE_AZ_4 = {'name': 'zone4', 'id': '9ca40ee4-3c2a-4635-9a18-233cf6e0ad0b'}
FAKE_AZ_5 = {'name': 'zone4', 'id': 'd76d921d-d6fa-41b4-a180-fb68952784bd'}
FAKE_AZ_6 = {'name': 'zone4', 'id': 'bc09c3d6-671c-4d55-9f43-f00757aabc50'}

SHARE_SERVICES_NO_POOLS = [
    dict(id=1, host='host1', topic='share', disabled=False,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_1['id'],
         availability_zone=FAKE_AZ_1),
    dict(id=2, host='host2@back1', topic='share', disabled=False,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_2['id'],
         availability_zone=FAKE_AZ_2),
    dict(id=3, host='host2@back2', topic='share', disabled=False,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_3['id'],
         availability_zone=FAKE_AZ_3),
]

SERVICE_STATES_NO_POOLS = {
    'host1': dict(share_backend_name='AAA',
                  total_capacity_gb=512, free_capacity_gb=200,
                  timestamp=None, reserved_percentage=0,
                  reserved_snapshot_percentage=0,
                  provisioned_capacity_gb=312,
                  max_over_subscription_ratio=1.0,
                  thin_provisioning=False,
                  snapshot_support=False,
                  create_share_from_snapshot_support=False,
                  revert_to_snapshot_support=True,
                  mount_snapshot_support=True,
                  driver_handles_share_servers=False),
    'host2@back1': dict(share_backend_name='BBB',
                        total_capacity_gb=256, free_capacity_gb=100,
                        timestamp=None, reserved_percentage=0,
                        reserved_snapshot_percentage=0,
                        provisioned_capacity_gb=400,
                        max_over_subscription_ratio=2.0,
                        thin_provisioning=True,
                        snapshot_support=True,
                        create_share_from_snapshot_support=True,
                        revert_to_snapshot_support=False,
                        mount_snapshot_support=False,
                        driver_handles_share_servers=False),
    'host2@back2': dict(share_backend_name='CCC',
                        total_capacity_gb=10000, free_capacity_gb=700,
                        timestamp=None, reserved_percentage=0,
                        reserved_snapshot_percentage=0,
                        provisioned_capacity_gb=50000,
                        max_over_subscription_ratio=20.0,
                        thin_provisioning=True,
                        snapshot_support=True,
                        create_share_from_snapshot_support=True,
                        revert_to_snapshot_support=False,
                        mount_snapshot_support=False,
                        driver_handles_share_servers=False),
}

SHARE_SERVICES_WITH_POOLS = [
    dict(id=1, host='host1@AAA', topic='share', disabled=False,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_1['id'],
         availability_zone=FAKE_AZ_1),
    dict(id=2, host='host2@BBB', topic='share', disabled=False,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_2['id'],
         availability_zone=FAKE_AZ_2),
    dict(id=3, host='host3@CCC', topic='share', disabled=False,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_3['id'],
         availability_zone=FAKE_AZ_3),
    dict(id=4, host='host4@DDD', topic='share', disabled=False,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_4['id'],
         availability_zone=FAKE_AZ_4),
    # service on host5 is disabled
    dict(id=5, host='host5@EEE', topic='share', disabled=True,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_5['id'],
         availability_zone=FAKE_AZ_5),
    dict(id=6, host='host6@FFF', topic='share', disabled=True,
         updated_at=timeutils.utcnow(), availability_zone_id=FAKE_AZ_6['id'],
         availability_zone=FAKE_AZ_6),
]

SHARE_SERVICE_STATES_WITH_POOLS = {
    'host1@AAA': dict(share_backend_name='AAA',
                      timestamp=None, reserved_percentage=0,
                      reserved_snapshot_percentage=0,
                      driver_handles_share_servers=False,
                      snapshot_support=True,
                      create_share_from_snapshot_support=True,
                      revert_to_snapshot_support=True,
                      replication_type=None,
                      pools=[dict(pool_name='pool1',
                                  total_capacity_gb=51,
                                  free_capacity_gb=41,
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=10,
                                  max_over_subscription_ratio=1.0,
                                  thin_provisioning=False)]),
    'host2@BBB': dict(share_backend_name='BBB',
                      timestamp=None, reserved_percentage=0,
                      reserved_snapshot_percentage=0,
                      driver_handles_share_servers=False,
                      snapshot_support=True,
                      create_share_from_snapshot_support=True,
                      revert_to_snapshot_support=False,
                      replication_type=None,
                      pools=[dict(pool_name='pool2',
                                  total_capacity_gb=52,
                                  free_capacity_gb=42,
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=60,
                                  max_over_subscription_ratio=2.0,
                                  thin_provisioning=True)]),
    'host3@CCC': dict(share_backend_name='CCC',
                      timestamp=None, reserved_percentage=0,
                      reserved_snapshot_percentage=0,
                      driver_handles_share_servers=False,
                      snapshot_support=True,
                      create_share_from_snapshot_support=True,
                      revert_to_snapshot_support=False,
                      replication_type=None,
                      pools=[dict(pool_name='pool3',
                                  total_capacity_gb=53,
                                  free_capacity_gb=43,
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=100,
                                  max_over_subscription_ratio=20.0,
                                  thin_provisioning=True)]),
    'host4@DDD': dict(share_backend_name='DDD',
                      timestamp=None, reserved_percentage=0,
                      reserved_snapshot_percentage=0,
                      driver_handles_share_servers=False,
                      snapshot_support=True,
                      create_share_from_snapshot_support=True,
                      revert_to_snapshot_support=False,
                      replication_type=None,
                      pools=[dict(pool_name='pool4a',
                                  total_capacity_gb=541,
                                  free_capacity_gb=441,
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=800,
                                  max_over_subscription_ratio=2.0,
                                  thin_provisioning=True),
                             dict(pool_name='pool4b',
                                  total_capacity_gb=542,
                                  free_capacity_gb=442,
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=2000,
                                  max_over_subscription_ratio=10.0,
                                  thin_provisioning=True)]),
    'host5@EEE': dict(share_backend_name='EEE',
                      timestamp=None, reserved_percentage=0,
                      reserved_snapshot_percentage=0,
                      driver_handles_share_servers=False,
                      snapshot_support=True,
                      create_share_from_snapshot_support=True,
                      revert_to_snapshot_support=False,
                      replication_type=None,
                      pools=[dict(pool_name='pool5a',
                                  total_capacity_gb=551,
                                  free_capacity_gb=451,
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=100,
                                  max_over_subscription_ratio=1.0,
                                  thin_provisioning=False),
                             dict(pool_name='pool5b',
                                  total_capacity_gb=552,
                                  free_capacity_gb=452,
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=100,
                                  max_over_subscription_ratio=1.0,
                                  thin_provisioning=False)]),
    'host6@FFF': dict(share_backend_name='FFF',
                      timestamp=None, reserved_percentage=0,
                      reserved_snapshot_percentage=0,
                      driver_handles_share_servers=False,
                      snapshot_support=True,
                      create_share_from_snapshot_support=True,
                      revert_to_snapshot_support=False,
                      replication_type=None,
                      pools=[dict(pool_name='pool6a',
                                  total_capacity_gb='unknown',
                                  free_capacity_gb='unknown',
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=100,
                                  max_over_subscription_ratio=1.0,
                                  thin_provisioning=False),
                             dict(pool_name='pool6b',
                                  total_capacity_gb='unknown',
                                  free_capacity_gb='unknown',
                                  reserved_percentage=0,
                                  reserved_snapshot_percentage=0,
                                  provisioned_capacity_gb=100,
                                  max_over_subscription_ratio=1.0,
                                  thin_provisioning=False)]),
}


class FakeFilterScheduler(filter.FilterScheduler):
    def __init__(self, *args, **kwargs):
        super(FakeFilterScheduler, self).__init__(*args, **kwargs)
        self.host_manager = host_manager.HostManager()


class FakeHostManager(host_manager.HostManager):
    def __init__(self):
        super(FakeHostManager, self).__init__()

        self.service_states = {
            'host1': {'total_capacity_gb': 1024,
                      'free_capacity_gb': 1024,
                      'allocated_capacity_gb': 0,
                      'thin_provisioning': False,
                      'reserved_percentage': 10,
                      'reserved_snapshot_percentage': 5,
                      'timestamp': None,
                      'snapshot_support': True,
                      'create_share_from_snapshot_support': True,
                      'replication_type': 'writable',
                      'replication_domain': 'endor',
                      'storage_protocol': 'NFS_CIFS',
                      },
            'host2': {'total_capacity_gb': 2048,
                      'free_capacity_gb': 300,
                      'allocated_capacity_gb': 1748,
                      'provisioned_capacity_gb': 1748,
                      'max_over_subscription_ratio': 2.0,
                      'thin_provisioning': True,
                      'reserved_percentage': 10,
                      'reserved_snapshot_percentage': 5,
                      'timestamp': None,
                      'snapshot_support': True,
                      'create_share_from_snapshot_support': True,
                      'replication_type': 'readable',
                      'replication_domain': 'kashyyyk',
                      'storage_protocol': 'NFS_CIFS',
                      },
            'host3': {'total_capacity_gb': 512,
                      'free_capacity_gb': 256,
                      'allocated_capacity_gb': 256,
                      'provisioned_capacity_gb': 256,
                      'max_over_subscription_ratio': 2.0,
                      'thin_provisioning': [False],
                      'reserved_percentage': 0,
                      'reserved_snapshot_percentage': 0,
                      'snapshot_support': True,
                      'create_share_from_snapshot_support': True,
                      'timestamp': None,
                      'storage_protocol': 'NFS_CIFS',
                      },
            'host4': {'total_capacity_gb': 2048,
                      'free_capacity_gb': 200,
                      'allocated_capacity_gb': 1848,
                      'provisioned_capacity_gb': 1848,
                      'max_over_subscription_ratio': 1.0,
                      'thin_provisioning': [True],
                      'reserved_percentage': 5,
                      'reserved_snapshot_percentage': 2,
                      'timestamp': None,
                      'snapshot_support': True,
                      'create_share_from_snapshot_support': True,
                      'replication_type': 'dr',
                      'replication_domain': 'naboo',
                      'storage_protocol': 'NFS_CIFS',
                      },
            'host5': {'total_capacity_gb': 2048,
                      'free_capacity_gb': 500,
                      'allocated_capacity_gb': 1548,
                      'provisioned_capacity_gb': 1548,
                      'max_over_subscription_ratio': 1.5,
                      'thin_provisioning': [True, False],
                      'reserved_percentage': 5,
                      'reserved_snapshot_percentage': 2,
                      'timestamp': None,
                      'snapshot_support': True,
                      'create_share_from_snapshot_support': True,
                      'replication_type': None,
                      'storage_protocol': 'NFS_CIFS',
                      },
            'host6': {'total_capacity_gb': 'unknown',
                      'free_capacity_gb': 'unknown',
                      'allocated_capacity_gb': 1548,
                      'thin_provisioning': False,
                      'reserved_percentage': 5,
                      'reserved_snapshot_percentage': 2,
                      'snapshot_support': True,
                      'create_share_from_snapshot_support': True,
                      'timestamp': None,
                      'storage_protocol': 'GLUSTERFS',
                      },
        }


class FakeHostState(host_manager.HostState):
    def __init__(self, host, attribute_dict):
        super(FakeHostState, self).__init__(host)
        for (key, val) in attribute_dict.items():
            setattr(self, key, val)


FAKE_HOST_STRING_1 = 'openstack@BackendA#PoolX'
FAKE_HOST_STRING_2 = 'openstack@BackendB#PoolY'
FAKE_HOST_STRING_3 = 'openstack@BackendC#PoolZ'


def mock_host_manager_db_calls(mock_obj, disabled=None):
    services = [
        dict(id=1, host='host1', topic='share', disabled=False,
             availability_zone=FAKE_AZ_1, availability_zone_id=FAKE_AZ_1['id'],
             updated_at=timeutils.utcnow()),
        dict(id=2, host='host2', topic='share', disabled=False,
             availability_zone=FAKE_AZ_1, availability_zone_id=FAKE_AZ_1['id'],
             updated_at=timeutils.utcnow()),
        dict(id=3, host='host3', topic='share', disabled=False,
             availability_zone=FAKE_AZ_2, availability_zone_id=FAKE_AZ_2['id'],
             updated_at=timeutils.utcnow()),
        dict(id=4, host='host4', topic='share', disabled=False,
             availability_zone=FAKE_AZ_3, availability_zone_id=FAKE_AZ_3['id'],
             updated_at=timeutils.utcnow()),
        dict(id=5, host='host5', topic='share', disabled=False,
             availability_zone=FAKE_AZ_3, availability_zone_id=FAKE_AZ_3['id'],
             updated_at=timeutils.utcnow()),
        dict(id=6, host='host6', topic='share', disabled=False,
             availability_zone=FAKE_AZ_4, availability_zone_id=FAKE_AZ_4['id'],
             updated_at=timeutils.utcnow()),
    ]
    if disabled is None:
        mock_obj.return_value = services
    else:
        mock_obj.return_value = [service for service in services
                                 if service['disabled'] == disabled]


class FakeWeigher1(base_host_weigher.BaseHostWeigher):
    def __init__(self):
        pass


class FakeWeigher2(base_host_weigher.BaseHostWeigher):
    def __init__(self):
        pass


class FakeClass(object):
    def __init__(self):
        pass


def fake_replica_request_spec(**kwargs):
    request_spec = {
        'share_properties': {
            'id': 'f0e4bb5e-65f0-11e5-9d70-feff819cdc9f',
            'name': 'fakename',
            'size': 1,
            'share_network_id': '4ccd5318-65f1-11e5-9d70-feff819cdc9f',
            'availability_zone': 'fake_az',
            'replication_type': 'dr',
        },
        'share_instance_properties': {
            'id': '8d5566df-1e83-4373-84b8-6f8153a0ac41',
            'share_id': 'f0e4bb5e-65f0-11e5-9d70-feff819cdc9f',
            'host': 'openstack@BackendZ#PoolA',
            'status': 'available',
            'availability_zone_id': 'f6e146d0-65f0-11e5-9d70-feff819cdc9f',
            'share_network_id': '4ccd5318-65f1-11e5-9d70-feff819cdc9f',
            'share_server_id': '53099868-65f1-11e5-9d70-feff819cdc9f',
        },
        'share_proto': 'nfs',
        'share_id': 'f0e4bb5e-65f0-11e5-9d70-feff819cdc9f',
        'snapshot_id': None,
        'share_type': 'fake_share_type',
        'share_group': None,
    }
    request_spec.update(kwargs)
    return request_spec


def get_fake_host(host_name=None):

    class FakeHost(object):
        def __init__(self, host_name=None):
            self.host = host_name or 'openstack@BackendZ#PoolA'

    class FakeWeightedHost(object):
        def __init__(self, host_name=None):
            self.obj = FakeHost(host_name=host_name)

    return FakeWeightedHost(host_name=host_name)
